package limone;

import java.util.*;
import eBeacon.*;
import limone.util.*;

/**
 * The Acquaintance List is a list of Profiles of the Agents within
 * communication range that this agent considers interesting.  It is one of
 * the objects inhereted when extending the agent.  It
 * allows the agent to obtain AgentProxy objects, which in turn is used by the
 * application's agent to interact with remote agents.
 *
 * <p> The Acquaintance List also allows objects to register as
 * AcquaintanceChangedListener
 * listeners.  By registering as a listener, the object is notified whenever
 * an agent is added or removed from the Acquaintance List.  This is useful
 * when the application wishes to create a "buddy list" consisting of the
 * agents with whom the user can interact with.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public class AcquaintanceList
{
    /**
	 * Maps AgentIDs to Profiles.
	 */
    Hashtable agentProfileTable = new Hashtable();
    
	/**
	 * A reference to the agent that this acquaintance list belongs to.
	 */
    Agent a;
	
	/**
	 * The AcquaintanceChangedListener listeners.
	 */
	Vector listeners = new Vector();
    
    /**
	 * Creates an Acquaintance List.
	 *
	 * @param a the agent that created and owns this Acquaintance List.
	 */
    public AcquaintanceList(Agent a) {
		this.a = a;
    }
	
	/**
	 * Adds a change listener to this acquaintance list.
	 */
	public void addChangeListener(AcquaintanceChangedListener acl) {
		listeners.addElement(acl);
	}
	
	/**
	 * Removes a change listener from this acquaintance list.
	 */
	public void removeChangeListener(AcquaintanceChangedListener acl) {
		listeners.removeElement(acl);
	}
    
    /**
	 * Determines whether this list contains a particular agent.
	 */
    public boolean contains(AgentID aID) {
		return agentProfileTable.containsKey(aID);
    }
	
	/**
	 * This is called when the engagement policy has changed.
	 */
	public void engPolicyChanged() {
		
		Enumeration enum = agentProfileTable.elements();
		Vector toRemove = new Vector();
		
		// for each profile within the acquaintance list
		while(enum.hasMoreElements()) {
			Profile p = (Profile)enum.nextElement();
			if (!a.getEP().isRelevant(p))
				toRemove.addElement(p);
		}
		
		for (int i = 0; i < toRemove.size(); i++) {
			profileRemoved((Profile)toRemove.elementAt(i));
		}
		
		// For each profile within the neighbor list that is not
		// already in the acquaintance list, check to see if they are
		// relevant.  If it is, add it to the acquaintance list.
		
		Enumeration neighbors = a.getNeighborList().elements();
		Vector toAdd = new Vector();
		
		while(neighbors.hasMoreElements()) {
			Profile p = (Profile)enum.nextElement();
			if (!agentProfileTable.contains(p) && a.getEP().isRelevant(p))
				toAdd.addElement(p);
		}
		
		for (int i = 0; i < toAdd.size(); i++) {
			profileAdded((Profile)toAdd.elementAt(i));
		}
		
	}
	
	/**
	 * Returns an array of profiles within this acquaintance list
	 * that match the specified profile selector.
	 *
	 * @param ps the profile selector to consider
	 * @return the profiles that match the specified profile selector.
	 */
	public Profile[] getAgentProfiles(ProfileSelector ps) {
		return getAgentProfiles(new ProfileSelector[] {ps});
	}
	
	/**
	 * Returns an array of profiles within this acquaintance list
	 * that match at least one of the specified profile selectors.
	 *
	 * @param pss the profile selectors to consider
	 * @return the profiles that match the specified profile selectors.
	 */
	public Profile[] getAgentProfiles(ProfileSelector[] pss) {
		
		Vector results = new Vector();
		Enumeration enum = agentProfileTable.elements();
		
		// for each profile within the acquaintance list
		while(enum.hasMoreElements()) {
			Profile p = (Profile)enum.nextElement();
			boolean matches = false;
			for (int i = 0; !matches && i < pss.length; i++) {
				matches = (!matches && pss[i].matches(p));
			}
			
			if (matches)
				results.addElement(p);
		}
		
		// convert the vector into an array of AgentLocations
		Profile[]  result = new Profile[results.size()];
		for (int i = 0; i < results.size(); i++)  {
			result[i] =(Profile)results.elementAt(i);
		}
		return result;
	}
	
    /**
	 * Returns an array of agent proxies of agents within this acquaintance
	 * list that match the specified profile selector.
	 *
	 * @param ps the profile selector to consider
	 * @return the agent proxies that match the specified profile selector.
	 */
    public AgentProxy[] getAgentProxies(ProfileSelector ps) {
		return getAgentProxies(new ProfileSelector[] {ps});
    }
	
    /**
	 * Returns an array of agent proxies of agents within this acquaintance
	 * list that match at least one of the specified profile selectors.
	 *
	 * @param ps the profile selector to consider
	 * @return the agent proxies that match at least one of the specified
	 * profile selectors.
	 */
    public AgentProxy[] getAgentProxies(ProfileSelector[] pss) {
		Profile[] profiles = getAgentProfiles(pss);
		
		// convert array of profiles into an array of agent proxies
		AgentProxy[]  result = new AgentProxy[profiles.length];
		for (int i = 0; i < profiles.length; i++)  {
			result[i] = new AgentProxy(profiles[i], a);
		}
		return result;
    }
    
    /**
	 * Returns the Profile of the Agent with AgentID aID.
	 * If the Agent is NOT in the acquaintance list, then null is returned.
	 *
	 * @param aID the AgentID of the agent whose profile we want to obtain.
	 * @return the Profile of the agent with the specified AgentID, or null
	 * if such an agent does not exist in the list.
	 */
    public Profile getAgentProfile(AgentID aID) {
		return (Profile)agentProfileTable.get(aID);
    }
	
	/**
	 * Returns a proxy for the agent with the specified agent id.
	 * If such an agent is not present, then null is returned.
	 *
	 * @param aID the AgentID of the agent whose proxy we want to obtain.
	 * @return a proxy to the agent with the specified agent ID, or null
	 * if such an agent does not exist in the list.
	 */
	public AgentProxy getAgentProxy(AgentID aID) {
		log("Getting proxy for agent " + aID);
		printList();
		
		Profile p = getAgentProfile(aID);
		
		if (p != null) {
			log("Got the proxy.");
			return new AgentProxy(p, a);
		} else {
			log("Could not get the proxy.");
			return null;
		}
	}
    
    /**
	 * Called by the neighbor list when the profile of an agent is added to
	 * to tho neighbor list.
	 *
	 * @param p the profile of the agent to be added to the list.
	 */
    void profileAdded(Profile p) {
		
		AgentID aID = ProfileParser.getAgentID(p);
		
		// if the agent is not already in the acquaintance list
		if (agentProfileTable.get(aID) == null) {
			
			agentProfileTable.put(aID, p);
			
			log("Added agent " + aID);
			printList();
			
			a.getRxnMgr().agentAdded(p);
			notifyListeners(new AcquaintanceChangedEvent(
								AcquaintanceChangedEvent.AGENT_ADDED, p));
		} else
			profileChanged(p);
    }
    
    /**
	 * Called by the neighbor list when the profile of an agent has changed.
	 *
	 * @param p the new profile of the agent.
	 */
    void profileChanged(Profile p) {
		
		log("profileChanged(Profile p): Called...");
		
		AgentID aID = ProfileParser.getAgentID(p);
		Profile oldp = (Profile)agentProfileTable.remove(aID);
		if (oldp != null) {
			a.getRxnMgr().agentRemoved(oldp);
			log("profileChanged(Profile p): Removed " + aID);
		}
		
		if (a.getEP().isRelevant(p)) {
			agentProfileTable.put(aID, p);
			
			log("Added " + aID);
			
			a.getRxnMgr().agentAdded(p);
			notifyListeners(new AcquaintanceChangedEvent(
								AcquaintanceChangedEvent.AGENT_ADDED, p));
		}
		
    }
    
    /**
	 * Called by the neighbor list when the profile of an agent has been
	 * removed due to loss of beacon reception.
	 *
	 * @param p the profile that was removed.
	 */
	void profileRemoved(Profile p) {
		
		AgentID aID = ProfileParser.getAgentID(p);
		
		if (agentProfileTable.remove(aID) != null) {
			
			log("profiledRemoved(Profile p): Removed profile" + aID);
			printList();
			
			a.getRxnMgr().agentRemoved(p);
			notifyListeners(new AcquaintanceChangedEvent(
								AcquaintanceChangedEvent.AGENT_REMOVED, p));
		}
    }
    
	/**
	 * Goes through each AcquaintanceChangedListener registered
	 * on this acquaintance list and notifies them that the acquaintance
	 * list has changed.
	 */
	private void notifyListeners(AcquaintanceChangedEvent ace) {
		for (int i = 0; i < listeners.size(); i++) {
			AcquaintanceChangedListener acl =
				(AcquaintanceChangedListener)listeners.elementAt(i);
			acl.acuaintanceChanged(ace);
		}
	}
    
    /**
	 * Returns a string representation of this class.
	 *
	 * @return a String representation of this class.
	 */
    public String toString() {
		String result = "Acquaintance list:";
		Enumeration enum = agentProfileTable.elements();
		while(enum.hasMoreElements()) {
			Profile p = (Profile)enum.nextElement();
			result += "\n\t" + p.toString();
		}
		return result;
    }
	
	/**
	 * Prints the contents of the acquaintance list.
	 */
	private void printList() {
		String result = "Contents of list:";
		
		Enumeration enum = agentProfileTable.keys();
	 	while(enum.hasMoreElements()) {
			AgentID aID = (AgentID)enum.nextElement();
			Profile p = (Profile)agentProfileTable.get(aID);
			result += "\n\t" + aID;//+ " - " + p;
		}
		log(result);
	
	}
	
	private void log(String msg) {
		DebugLogger.log("AcquaintanceList", msg);
	}
}

