package limone;

import eLights.*;
import eBeacon.*;
import java.util.*;
import java.io.*;
import mucode.*;
import mucode.abstractions.*;
import limone.io.*;

/**
 * The Agent is the unit of execution and modularity in Limone.
 * As such, all applications written on top of Limone must extend this agent.
 * That is, in order to use Limone, the application developer creates an
 * object that extends this class, then augments her class with the code
 * associated with the application.  This is done by implementing
 * the abstract methods defined within this agent such as agentInit(),
 * agentRun(), agentDone().
 *
 * <p> The agent has two default properties in its profile:
 * <ol>
 * <li>"_AgentID" - the agent's AgentID</li>
 * <li>"_HostID" - the HostID of the LimoneServer on which the agent is
 *     located </li>
 * </ol>
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public abstract class Agent extends MuAgent implements Serializable
{
	/**
	 * The user-defined name of this agent.
	 */
	String name;
	
	/**
	 * The AgentID uniquely identifies this agent.
	 * It is assigned by the LimeLiteServer when it
	 * loads the Agent.
	 */
    AgentID aID;
	
	/**
	 * True if this agent should beacon its profile.
	 * This should be changed to false prior to migration.
	 * By default it is true.
	 */
	boolean shouldBeacon = true;
	
	/**
	 * This is the number of milliseconds that can pass since receiving
	 * the last beacon from a particular agent before considering the agent is
	 * gone.
	 */
	protected long disengagement_threshold = 3000;
	
	/**
	 * This is the amount of time in ms an agent will wait for a reply while
	 * performing a remote operation before it times out.
	 */
	protected long remote_operation_timeout = 2000;
	
	/**
	 * This agent's local tuple space.
	 */
	protected TupleSpace tupleSpace;
	
	/**
	 * The reaction registry stores all of the reactions registered by this
	 * agent.
	 */
	protected ReactionRegistry rr;
	
	/**
	 * Responsible for propagating and un-propagating reactions, and executing
	 * the callback function when a reaction fires.
	 */
	private ReactionManager rxnMgr;
	
	/**
	 * The engagement policy determines which neighboring agents are relevant.
	 */
	protected EngagementPolicy engPolicy;
	
	/**
	 * The acquaintance list keeps track of the relevant neighbors.
	 */
	protected AcquaintanceList aql;
	
	/**
	 * The neighbor list contains all of the neighboring agents
	 * within communication range.
	 */
	private NeighborList nl;
	
	/**
	 * The profile of this agent.  This profile describes properties of this
	 * agent and is continuously broadcasted.  The two default properties
	 * within this profile are [_AgentID, AgentID] and [_HostID, HostID].
	 */
	protected Profile profile;
	
	/**
	 * The operation manager controlls whether a remote operation is
	 * accepted.
	 */
	protected OperationManager opMgr;
	
	/**
	 * This is the LimoneServer on top of which this Agent runs.
	 */
    private LimoneServer ls;
	
	/**
	 * The reaction list contains the reactions that apply to the local
	 * tuple space.
	 */
	private ReactionList rl;
	
	/**
	 * The remote operation manager handles the remote operations that
	 * this agent has requested.
	 */
	private RemoteOpManager remoteOpMgr;
	
	/**
	 * The RemoteRequestManager is responsible for handling remote
	 * requests.
	 */
	private RemoteRequestManager remoteRequestMgr;
	/**
	 * The beacon manager is responsible for managing the contents of the
	 * neighbor list.
	 */
	private BeaconManager beaconMgr;
	
	/**
	 * This is the thread that executes the AgentCode method.
	 */
	private AgentThread agentThread;
	
	/**
	 * This default constructor is necessary for migration.
	 */
	public Agent() {
	}
	
	/**
	 * Creates an Agent.
	 *
	 * @param aID the AgentID of this agent.
	 * @param name the user-defined name of this agent
	 */
	public Agent(AgentID aID, String name) {
		this.aID = aID;
		this.name = name;
		
		this.ls = LimoneServer.getServer();
		
		profile = new Profile();
		profile.addProperty(new Property("_AgentID", aID));
		profile.addProperty(new Property("_HostID", ls.getHostID()));
		
		aql = new AcquaintanceList(this);
		rl = new ReactionList(this);
		rxnMgr = new ReactionManager(this);
		
		tupleSpace = new TupleSpace(this);
		opMgr = new DefaultOpManager();
		remoteRequestMgr = new RemoteRequestManager(this);
		remoteOpMgr = new RemoteOpManager(this);
		rr = new ReactionRegistry(this);
		nl = new NeighborList(this);
		
		engPolicy = new EngagementPolicy(this);
		
		beaconMgr = new BeaconManager(this);
	}
	
	/**
	 * Returns this agent's AgentID.  This is a globally unique ID that the
	 * agent has throughout its lifetime, even during migration.
	 */
	public AgentID getAgentID() {
		return aID;
	}
	
	/**
	 * Returns the HostID of the agent that this agent is residing on.
	 */
	public HostID getHostID() {
		return ls.getHostID();
	}
	
	/**
	 * An accessor to the acquaintance list.
	 */
	public AcquaintanceList getAQL() {
		return aql;
	}
	
	/**
	 * An accessor to the neighbor list.
	 */
	NeighborList getNeighborList() {
		return nl;
	}
	
	/**
	 * An accessor to the reaction manager.
	 */
	ReactionManager getRxnMgr() {
		return rxnMgr;
	}
	
	/**
	 * An accessor to the reaction manager.
	 */
	RemoteOpManager getRemoteOpMgr() {
		return remoteOpMgr;
	}
	
	/**
	 * Returns the number of milliseconds that agent will wait for a reply
	 * before timing out.
	 */
	long getOperationTimeout() {
		return remote_operation_timeout;
	}
	
	/**
	 * An accessor to the reaction registry.
	 */
	public ReactionRegistry getRR() {
		return rr;
	}
	
	/**
	 * An accessor to the reaction list.
	 */
	ReactionList getRL() {
		return rl;
	}
	
	/**
	 * An accessor to the tuple space.
	 */
	TupleSpace getTupleSpace() {
		return tupleSpace;
	}
	
	/**
	 * Used by the neighborList to determine when
	 * a remote host should be considered disengaged.
	 */
	long getDisengagementThreshold() {
		return disengagement_threshold;
	}
	
	/**
	 * This is called by the LimoneServer each time it receives a beacon.
	 *
	 * @param beacon the beacon that was just received
	 */
	void newBeacon(LimoneBeacon beacon) {
		beaconMgr.newBeacon(beacon);
	}
	
	/**
	 * Called by the LimoneServer when a new message destined for this
	 * agent has arrived.
	 */
	void newMsg(Message msg) {
		log("Received a new message");
		if (msg instanceof ReactionResults) {
			log("the message was a reaction results");
			rxnMgr.processRxnResults((ReactionResults)msg);
		} else if (msg instanceof Reply) {
			log("the message was a reply");
			remoteOpMgr.processReply((Reply)msg);
		} else if (msg instanceof Request) {
			log("the message was a request");
			remoteRequestMgr.processRequest((Request)msg);
		}
	}
	
	/**
	 * Sets the intial profiles to be located within the neighbor list.
	 */
	void setInitEng(Profile[] profiles) {
		for (int i = 0; i < profiles.length; i++) {
			nl.addProfile(profiles[i]);
		}
	}
	
	/**
	 * Returns true if this agent's profile should be beaconed.
	 */
	boolean includeInBeacons() {
		return shouldBeacon;
	}
	
	/**
	 * Sets the operation manager used by this agent.
	 */
	protected void setOperationMgr(OperationManager opmgr) {
		this.opMgr = opMgr;
	}
	
	/**
	 * Called when the LimoneServer on which this agent is residing is
	 * about to shut down.
	 */
	protected abstract void shutDown();
	
	/**
	 * An accessor to the engagement policy.
	 */
	public EngagementPolicy getEP() {
		return engPolicy;
	}
	
	/**
	 * Returns this agent's profile.
	 */
	public Profile getProfile() {
		return profile;
	}
	
	public void run() {
	}
	
	/**
	 * This is the method that the application inplements.
	 */
	public abstract void agentRun();
	
	/**
	 * This is executed when the agent is initialized.
	 */
	public abstract void agentInit();
	
	
	/**
	 * This is code that is executed when the agent is done.
	 */
	public abstract void agentDone();
	
	/**
	 * Migrates this agent to the specified host.
	 */
	public void agentMigrate(HostID hID) {
		//..
	}
	
	
	
	/**
	 * Returns a string representation of this class.
	 *
	 * @return a String representation of this class.
	 */
    public String toString() {
		StringBuffer sBuff = new StringBuffer("Agent ");
		sBuff.append(aID);
		return sBuff.toString();
    }

	void log(String msg) {
		DebugLogger.log("Agent " + getAgentID(), msg);
	}
}


