package limone;

/**
 * The AgentID is a globally unique identifier of an agent.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class AgentID implements java.io.Serializable{

    /**
     * A user-defineable name.  This is arbitrary.
     */
    final String name;
    
    /**
     * The ID of the Host.
     */
    final HostID creationHost;

    /**
     * An enumeration value of the AgentID.
     */
    final long enum;

    /**
     * Creates a new AgentID with the specified agentHost,
     * enumeration and name.  The value of the name is used
     * purely by the application developer, it is not used internally
     * by Limone.
     *
     * @param creationHost The HostID of the host that the agent
     *   represented by this AgentID is being loaded on.
     * @param enum the enumerated value provided by the Host.
     *   This value must be unique for each AgentID created by a host for
     *   a specific Agent.
     * @param name The name of the agent represented by this AgentID.
     *
     */
    public AgentID(HostID creationHost, long enum, String name) {
        this.creationHost = creationHost;
        this.enum = enum;
        this.name = name;
    }

    /**
     * Returns the user-definable name of the agent represented
     * by this AgentID.  This value is not used by Limone,
     * it is given merely as a convenience for application developers.
     *
     * @return the user-definable name of the agent represented
     * by this AgentID, or "anonymous" if no name was specified.
     */
    public String getName() {
        return name;
    }

    /**
     * Determines if another object is equal to this one.
     * Another object is equal to this one with the other object
     * is of type AgentID, both have the same enumeration value.
     */
    public boolean equals(Object o) {
        if (o != null && o instanceof AgentID) {
            AgentID aID = (AgentID)o;
            return aID.creationHost.equals(creationHost)
                && aID.enum == enum;
        } else
            return false;
    }

    /**
     * Returns the hashcode of this object.  This enables this object to
     * be used as a key in Hashtables.
     *
     * @return the hashcode of this object.
     */
    public int hashCode() {
        Integer code = new Integer((int)(creationHost.hashCode() + enum));
        return code.hashCode();
    }

    /**
     * Returns a String representation of this class.
     *
     * @return a String representation of this class.
     */
    public String toString() {
        StringBuffer sBuff = new StringBuffer("[");
        sBuff.append(creationHost.toString());
        sBuff.append(":a");
        sBuff.append(enum);
        sBuff.append(":");
        sBuff.append(name);
        sBuff.append("]");
        return sBuff.toString();
    }
}
