package limone;

import eLights.*;
import eBeacon.*;
import limone.util.*;


/**
 * An AgentProxy is the means by which an agent can request that another
 * agent perform an operation for it.  It is essentially a "proxy" to a
 * remote agent.  An AgentProxy is obtained using the acquaintance list
 * that is inherited when the application developer extends agent.
 *
 * <p>All operations done on an AgentProxy are simply requests for an
 * operation to be done.  The agent that the proxy represents may accept
 * or reject the request.  If the remote agent rejects the reqest, a
 * RejectionException is thrown.  Also, to account for message loss,
 * a timer is set immediately after sending the request.  If the timer runs
 * out prior to receiving a response, it is assumed that the messages
 * were lost during transmission and TimeOutException is thrown.
 *
 * @see Agent
 * @see AcquaintanceList
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public class AgentProxy
{
	/**
	 * This is the profile of the agent that this proxy represents.
	 */
	Profile p;
	
	
	/**
	 * The agent that this proxy belongs to.
	 */
	Agent a;
	
	/**
	 * The AgentID of the agent that this proxy represents.
	 */
	AgentID dAgentID;
	
	/**
	 * The HostID of the agent that this proxy represents.
	 */
	HostID dHostID;
	
	/**
	 * A reference to the remote operation manager.
	 */
	RemoteOpManager rmtOpMgr;
	
	/**
	 * Creates an AgentProxy for the agent that is represented by the profile
	 *
	 * @param p the profile of the agent that this proxy represents
	 * @param iom the IO manager that handles the sending and receving of
	 * messages.
	 */
	public AgentProxy(Profile p, Agent a) {
		this.p = p;
		this.a = a;
		rmtOpMgr = a.getRemoteOpMgr();
		dHostID = ProfileParser.getHostID(p);
		dAgentID = ProfileParser.getAgentID(p);
	}
	
	/**
	 * Sends a request to the agent represented by this proxy asking it to
	 * place a tuple into its tuple space.
	 *
	 * @param t the tuple to insert
	 * @throws RejectionException if the agent rejects the operation
	 * @throws TimeOutException if the operation timed out.
	 */
	public void out(ETuple t) throws RejectionException, TimeOutException {
						
		LocalTask lt = new LocalTask(LocalTask.OUT, dHostID, dAgentID, t);
		
		log("out(..): Created local task:\n\t" + lt);
		
		rmtOpMgr.addTask(lt);
			
		lt.holdUntilDone();  // wait until the operation actually completed
				
		// Check to see if the operation timed out or was rejected, throw
		// exceptions as necessary
		if (lt.timedOut)
			throw new TimeOutException(
				"The OUT operation to " + dAgentID + " timed out");
		else if (lt.rejected)
			throw new RejectionException(
				"The OUT operation to " + dAgentID + " was rejected.");
	}
	
	
	/**
	 * Sends a request to the agent represented by this proxy asking it to
	 * perform a probing INp operationg using the specified template.
	 *
	 * @param t the template to use in the INp operation
	 * @return the tuple that was obtained from the probing INp operation
	 * @throws RejectionException if the agent rejects the operation
	 * @throws TimeOutException if the operation timed out.
	 */
	public ETuple inp(ETemplate t) throws RejectionException, TimeOutException {
		LocalTask lt = new LocalTask(LocalTask.INP, dHostID, dAgentID, t);
		rmtOpMgr.addTask(lt);
		
		lt.holdUntilDone();  // wait until the operation actually completed
		
		// Check to see if the operation timed out or was rejected, throw
		// exceptions as necessary
		if (lt.timedOut)
			throw new TimeOutException(
				"The INp operation to " + dAgentID + " timed out");
		else if (lt.rejected)
			throw new RejectionException(
				"The INp operation to " + dAgentID + " was rejected.");
		else
			return lt.tuple;
	}
	
	/**
	 * Sends a request to the agent represented by this proxy asking it to
	 * perform a probing INpg operationg using the specified template.
	 *
	 * @param t the template to use in the INpg operation
	 * @return the tuples that were obtained from the probing INgp operation
	 * @throws RejectionException if the agent rejects the operation
	 * @throws TimeOutException if the operation timed out.
	 */
	public ETuple[] ingp(ETemplate t) throws RejectionException, TimeOutException {
		LocalTask lt = new LocalTask(LocalTask.INGP, dHostID, dAgentID, t);
		rmtOpMgr.addTask(lt);
		
		lt.holdUntilDone();  // wait until the operation actually completed
		
		// Check to see if the operation timed out or was rejected, throw
		// exceptions as necessary
		if (lt.timedOut)
			throw new TimeOutException(
				"The INgp operation to " + dAgentID + " timed out");
		else if (lt.rejected)
			throw new RejectionException(
				"The INgp operation to " + dAgentID + " was rejected.");
		else
			return lt.tuples;
	}
	
	/**
	 * Sends a request to the agent represented by this proxy asking it to
	 * perform a probing RDp operationg using the specified template.
	 *
	 * @param t the template to use in the RDp operation
	 * @return the tuple that was obtained from the probing RDp operation
	 * @throws RejectionException if the agent rejects the operation
	 * @throws TimeOutException if the operation timed out.
	 */
	public ETuple rdp(ETemplate t) throws RejectionException, TimeOutException {
		LocalTask lt = new LocalTask(LocalTask.RDP, dHostID, dAgentID, t);
		rmtOpMgr.addTask(lt);
		
		lt.holdUntilDone();  // wait until the operation actually completed
		
		// Check to see if the operation timed out or was rejected, throw
		// exceptions as necessary
		if (lt.timedOut)
			throw new TimeOutException(
				"The RDp operation to " + dAgentID + " timed out");
		else if (lt.rejected)
			throw new RejectionException(
				"The RDp operation to " + dAgentID + " was rejected.");
		else
			return lt.tuple;
	}
	
	/**
	 * Sends a request to the agent represented by this proxy asking it to
	 * perform a probing RDpg operationg using the specified template.
	 *
	 * @param t the template to use in the RDpg operation
	 * @return the tuples that were obtained from the probing RDgp operation
	 * @throws RejectionException if the agent rejects the operation
	 * @throws TimeOutException if the operation timed out.
	 */
	public ETuple[] rdgp(ETemplate t) throws RejectionException, TimeOutException {
		LocalTask lt = new LocalTask(LocalTask.RDGP, dHostID, dAgentID, t);
		rmtOpMgr.addTask(lt);
		
		lt.holdUntilDone();  // wait until the operation actually completed
		
		// Check to see if the operation timed out or was rejected, throw
		// exceptions as necessary
		if (lt.timedOut)
			throw new TimeOutException(
				"The RDgp operation to " + dAgentID + " timed out");
		else if (lt.rejected)
			throw new RejectionException(
				"The RDgp operation to " + dAgentID + " was rejected.");
		else
			return lt.tuples;
	}
	
	public String toString() {
		return "AgentProxy" + dAgentID;
	}
	
	private void log(String msg) {
		DebugLogger.log("AgentProxy " + dAgentID + ": ", msg);
	}
}


