package limone;

import java.util.*;
import java.io.*;
import java.net.*;

import limone.io.*;
import eBeacon.*;
import eLights.*;
import mucode.*;

import java.lang.reflect.*;

/**
 * The LimoneServer is the foundational process that provide the basis
 * on which agents may operate.  An application developer uses Limone by
 * extending an agent and augmenting it with its code.  The application is
 * run by loading the application's agent onto a LimoneServer using the
 * loadAgent(...) method.
 *
 * <p>The LimoneServer is a singleton.  It is accessed using
 * LimoneServer.getServer().  It should be launched by executing
 * limone.util.LimoneLauncher with the appropriate flags.
 *
 * <p>If the need arises, a LimoneServer can be booted manually from
 * within an application by calling boot on the LimoneServer.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public class LimoneServer implements PropertyKeys, BeaconListener, Serializable
{
	/**
	 * Use TCP for single-cast communication.
	 */
    public static final int TCP = 0;
    
    /**
	 * Use UDP for single-cast communication.
	 */
    public static final int UDP = 1;
    
	/**
	 * Encapsulates the properties of this Limone Server.
	 */
	PropertyVars pv = new PropertyVars();
	
    /**
	 * The ID of this server.
	 */
    HostID hostID;
	
    /**
	 * A reference to the LimoneServer singleton.
	 */
    static final LimoneServer server = new LimoneServer();
	
    /**
	 * Keeps track of whether this server has been booted.
	 */
    boolean booted = false;
    
    /**
	 * The number of agents running on top of this server.  It is used
	 * to when creating a new AgentID.
	 */
    long numAgents = 0;
    
    /**
	 * References all agents running on top of this server, keyed by AgentID.
	 */
    Hashtable agentTable = new Hashtable();
    
    /**
	 * The BeaconServer used by this LimoneServer.
	 */
    BeaconServer beaconServer;
    
    /**
	 * Sends messages to other agents.
	 */
    MessageSender msgSender;
    
    /**
	 * Listens for incomming messages on the unicast port.
	 */
    MessageReceiver msgReceiver;
    
    /**
	 * This is the muserver that handles the migration of an agent from
	 * one LimoneServer to another.
	 */
    MuServer muserver = null;
    
    /**
	 * Constructs a LimoneServer.
	 */
    private LimoneServer() {
    }
    
    /**
	 * Sets the user definable variables.  It takes String array of the form
	 * {key, value, key, value, ..., key, value}.  The available keys are:
	 *
	 * <ul>
	 * <li>-debug -makes limone run in debug mode</li>
	 * <li>-mPort -the multicast port number</li>
	 * <li>-sPort - the singlecast port number</li>
	 * <li>-mcastaddr - the multicast group address</li>
	 * <li>-protocol - either TCP or UDP</li>
	 * <li>-beaconPeriod - the period in ms between beacon broadcasting</li>
	 * </ul>
	 */
    public void parseArgs(String[] args)
		throws LimoneServerException{
		
		try {
			int i = 0;
			while (i < args.length) {
				
				if (args[i].equals("-" + DEBUG_KEY))
					pv.debug = true;
				else if (args[i].equals("-" + MUTICAST_PORT_KEY))
					pv.mcastport = Integer.valueOf(args[++i]).intValue();
				else if (args[i].equals("-" + SINGLECAST_PORT_KEY))
					pv.port = Integer.valueOf(args[++i]).intValue();
				else if (args[i].equals("-" + MULTICAST_ADDRESS_KEY))
					pv.mcastAddressString = args[++i];
				else if (args[i].equals("-" + PROTOCOL_KEY))
					pv.useTCP = args[++i].toLowerCase().equals("tcp");
				else if (args[i].equals("-" + BEACON_PERIOD_KEY))
					pv.beaconPeriod = Long.valueOf(args[++i]).longValue();
				i++;
			}
		} catch(Exception e) {
			throw new LimoneServerException(e);
		}
	}
	
	
	/**
	 * Sets the properties values.
	 */
	/*private void loadPropertyVars() throws LimoneServerException{
	 try {
	 
	 debugMode = Integer.valueOf(
	 properties.getProperty(DEBUG_KEY)).intValue();
	 
	 port = Integer.valueOf(
	 properties.getProperty(SINGLECAST_PORT_KEY)).intValue();
	 
	 mcastAddressString = properties.getProperty(MULTICAST_ADDRESS_KEY);
	 
	 mcastport = Integer.valueOf(
	 properties.getProperty(MUTICAST_PORT_KEY)).intValue();
	 
	 useTCP = properties.getProperty(PROTOCOL_KEY).equals("TCP");
	 
	 beaconPeriod = Long.valueOf(
	 properties.getProperty(BEACON_PERIOD_KEY)).longValue();
	 } catch (Exception e) {
	 e.printStackTrace();
	 throw new LimoneServerException("Error loading property vars.");
	 }
	 }*/
	
	/**
	 * Activates the LimoneServer.  Once it is activated, neither the uni-cast
	 * nor multicast port can be changed.  If the user tries to boot a LimoneServer
	 * that is already booted, then a LimoneServerException will be thrown.
	 *
	 * <p> A LimoneServerException is also thrown if the unicast port
	 * has already in use.
	 *
	 */
	public synchronized void boot() throws LimoneServerException {
		
		if (booted)
			throw new LimoneServerException("LimoneServer already booted");
		else {
			
			// set the ID of this server
			try{
				hostID = new HostID(InetAddress.getLocalHost(), pv.port);
			} catch(UnknownHostException e) {
				log("Error creating the host id");
				e.printStackTrace();
				System.exit(1);
			}
			
			log("The hostID is " + hostID);
			
			// load the number of agents this LimoneServer has created before
			StringBuffer fileName = new StringBuffer("lls");
			fileName.append(hostID.toString()).append(".dat");
			boolean fileExists = true;
			FileReader fr = null;
			try{
				fr = new FileReader(fileName.toString());
			} catch(FileNotFoundException exc) {
				fileExists = false;
			}
			if (fileExists) {
				BufferedReader in = new BufferedReader(fr);
				try{
					numAgents = new Long(in.readLine()).longValue();
				} catch(IOException ioe) {
					numAgents = 0;
				} catch(NumberFormatException e) {
					numAgents = 0;
				}
			} else
				numAgents = 0;
			
			// initialize the muServer
			muserver = new MuServer();
			muserver.addUbiquitousPackage("limone");
			muserver.addUbiquitousPackage("eLights");
			muserver.addUbiquitousPackage("eBeacon");
			muserver.addUbiquitousPackage("mucode");
			
			String[] mucodeArgs =
				new String[] {"-port", Integer.toString(pv.port+1)};
			
			new mucode.util.Launcher(muserver).parseArgs(mucodeArgs, 0);
			muserver.boot();
			
			//create the message sender and receiver
			if (pv.useTCP) {
				msgSender = new TCPMessageSender();
				msgReceiver = new TCPMessageReceiver(pv.port);
			}
			else {
				msgSender = new UDPMessageSender();
				msgReceiver = new UDPMessageReceiver(pv.port);
			}
			
			// obtain the multicast address
			InetAddress mcastGroupAddress = null;
			try {
				mcastGroupAddress =
					InetAddress.getByName(pv.mcastAddressString);
			}
			catch (UnknownHostException uhe) {
				log("Problems getting multicast address");
				uhe.printStackTrace();
				System.exit(1);
			}
			
			// create the BeaconReceiver
			BeaconReceiver br =
				new BeaconReceiver(mcastGroupAddress, pv.mcastport);
			
			// create the BeaconServer
			beaconServer = new BeaconServer(br, createBeacon());
			
			beaconServer.setPeriod(pv.beaconPeriod);
			
			// add this LimoneServer as a listener
			beaconServer.addBeaconListener(this);
			
			// start the BeaconServer
			beaconServer.start();
			
			// enable the LimoneServer
			booted = true;
			log("The LimoneServer has been booted.");
		}
	}
	
	/**
	 * Returns true if this server is booted.
	 */
	public boolean isBooted() {
		return booted;
	}
	
	/**
	 * Creates and returns the bytes of the beacon to be broadcasted.
	 * The beacon consists of a list of Agent profiles.  All agents
	 * MUST have a profile!
	 *
	 * @return a serialized version of the beacon to be broadcasted.
	 */
	private LimoneBeacon createBeacon() {
		
		LimoneBeacon beacon = null;
		
		if (agentTable.size() == 0)
			beacon = new LimoneBeacon(msgSender.getProtocolType(),hostID);
		else {
			Vector agentVector = new Vector();
			
			// Get the agents that should be included in the beacon
			Enumeration enum = agentTable.elements();
			
			while(enum.hasMoreElements()) {
				
				Agent agent = (Agent)enum.nextElement();
				
				if (agent.includeInBeacons())
					agentVector.addElement(agent);
				
			}
			
			// Create an array that holds their location
			Profile[] profiles = new Profile[agentVector.size()];
			
			log("CreateBeacon: number of profiles: " + profiles.length);
			
			for (int i = 0; i < agentVector.size(); i++) {
				Agent agent = (Agent)agentVector.elementAt(i);
				profiles[i] = agent.getProfile();
			}
			
			// create the beacon
			beacon = new LimoneBeacon(msgSender.getProtocolType(),
									  hostID, profiles);
		}
		
		return beacon;
	}
	
	/**
	 * Updates the current value of the beacons to be broadcasted
	 * by the BeaconBroadcaster.
	 */
	protected synchronized void refreshBeacon() {
		beaconServer.setBeacon(createBeacon());
	}
	
	/**
	 * Shuts down an Agent.  The Agent being shut down must
	 * currently be running on top of this LimoneServer.  This method shuts
	 * an agent down by removing it from the agent table and updating
	 * the beacon.  Since the agent was removed from the agent table, it
	 * will no longer be included with the beacons.  Thus, it would appear
	 * to no longer exist from other agent's perspective.
	 *
	 * @param aID the AgentID of the Agent being shutdown.
	 */
	public void killAgent(AgentID aID) {
		agentTable.remove(aID);
		refreshBeacon();
	}
	
	/**
	 * Shutdowns this LimoneServer and exits the system.
	 */
	public void shutdown() {
		shutdown(true);
	}
	
	/**
	 * Kills this LimeServer.  Disables the beaconing mechanism and
	 * saves the number of Agents created.  This does not shutdown the
	 * java virtual machine.
	 *
	 * @param exitSystem true if the JVM should be exited at the end of
	 * this method call.
	 */
	public void shutdown(boolean exitSystem) {
		log("Shutdown called");
		
		if (booted) {
			
			log("Killing each agent on the server");
			
			// kill each agent running on top of this server
			Enumeration agentEnum = agentTable.elements();
			
			while(agentEnum.hasMoreElements()) {
				((Agent)agentEnum.nextElement()).shutDown();
			}
			
			// kill the beacons
			log("shutting down the beacon server.");
			beaconServer.stop();
			
			
			// kill the message sender and receiver
			log("killing the message sender.");
			msgSender.kill();
			
			log("killing the message receiver.");
			msgReceiver.kill();
			
			// save the number of agents created
			log("Saving the number of agents created.");
			StringBuffer fileName = new StringBuffer("lls");
			fileName.append(hostID.toString()).append(".dat");
			
			log("Saving settings in file: " + fileName.toString());
			FileOutputStream fos = null;
			try{
				fos = new FileOutputStream(fileName.toString(), false);
				String number = (new Long(numAgents)).toString();
				fos.write(number.getBytes());
			} catch(IOException ioe) {
				ioe.printStackTrace();
				System.exit(1);
			} finally {
				try {
					fos.close();
				} catch(Exception e) {}
			}
			booted = false;
		}
		
		log("Server has been shutdown.");
		
		if (exitSystem) System.exit(0);
	}
	
	/**
	 * Returns the number of agents running on top of
	 * this LimoneServer.
	 *
	 * @return the number of agents running on top of
	 * this LimoneServer.
	 */
	public int numAgents() {
		return agentTable.size();
	}
	
	
	/**
	 * Return a handle to the Lime server.   The LimoneServer is a
	 * singleton.  It does not have to be instantiated.  Before agents can be
	 * loaded onto the LimoneServer, it must first be booted.
	 *
	 * @return a handle to the Lime server.
	 */
	public static LimoneServer getServer() {
		return server;
	}
	
	/**
	 * Returns this Limone server's muserver.
	 */
	public MuServer getMuServer() {
		return muserver;
	}
	
	/**
	 * Return the ID of this server, or null of the server has not booted yet.
	 *
	 * @return the unique <code>HostID</code> of this host
	 */
	public HostID getHostID() {
		return hostID;
	}
	
	/**
	 * Load a named agent into this LimoneServer.
	 *
	 * <p>The agent's class is described by a
	 * fully specified class name. The bytecode (<code>.class</code> files) for
	 * this class and all the other classes needed by the agent must be in a
	 * directory in the same <code>CLASSPATH</code> available to the Limone
	 * server.
	 *
	 * @param classname the full-path class name of the agent to be loaded.
	 * This class must extend Limone.Agent.
	 * @param agentName the user-defined name of the agent.
	 * @return the agent that was loaded onto this LimoneServer.
	 */
	public synchronized Agent loadAgent(String className, String agentName)
		throws AgentCreationException, LimoneServerException {
		System.out.println("loadAgent is started");
		Agent agent = null;
		
		// first make sure the LimoneServer is booted
		if (!booted)
			throw new LimoneServerException(
				"Limone server not booted yet, "
					+ "cannot load an agent until the server is booted.");
		else {
			
			logLoadAgent(
				"\n----------------------------------------"
					+ "\nLoading an agent!!"
					+ "\n\tclassName: " + className
					+ "\n\tagentName: " + agentName
					+ "\n----------------------------------------");
			
			// Calculate the name of the agent.
			String aName = agentName;
			if (aName == null || aName.equals(""))
				aName = "Anonymous" + numAgents;
			
			// create the agent ID
			AgentID aID = new AgentID(hostID, numAgents, aName);
			logLoadAgent("The AgentID is: " + aID);
			
			logLoadAgent("AgentID: " + aID);
			
			// get the agent's class
			Class agentClass = null;
			try {
				agentClass = Class.forName(className);
			} catch(ClassNotFoundException cnfe) {
				throw new AgentCreationException(
					cnfe,"Impossible to find some of the agent classes.");
			}
			
			// Make sure agent is a subclass of Limone.Agent.
			if (!(Agent.class).isAssignableFrom(agentClass))
				throw new AgentCreationException(
					"Agent creation aborted: The class supplied is not"
						+ " a subclass of limone.Agent.  All agents must"
						+ " subclass limone.Agent!!");
			
			logLoadAgent("Getting the constructor...");
			Class[] paramTypes = new Class[] {AgentID.class, String.class};
			Constructor constr = null;
			try {
				constr = agentClass.getConstructor(paramTypes);
			} catch(Exception e) {
				throw new AgentCreationException(
					e,"Could not get the agent's consructor");
			}
			
			logLoadAgent("Creating the agent...");
			// create the agent
			try {
				agent = (Agent) constr.newInstance(new Object[] {aID, aName});
			} catch(Exception e) {
				throw new AgentCreationException(
					e, "Could not create a new instance of the agent");
			}
			
			// Add the agent to the agentTable if it is not null
			System.out.println("loadAgent assert agent != null is " + (agent != null));
			if (agent != null) {
				logLoadAgent("Adding agent to agentTable");
				
				// update the number of agents running on this host
				numAgents++;
				
				// add the new agent into the hashtable
				agentTable.put(agent.getAgentID(), agent);
				
				// tell the agent to initially engage with all local agents
				agent.setInitEng(getAgentProfiles());
				
				// refresh the beacon to be broadcasted
				refreshBeacon();
				
				// create a new thread for the agent.
				new AgentThread(agent);
				
				logLoadAgent(
					"\n----------------------------------------"
						+ "\n\n\nAgent " + className
						+ ", " + aID + ", loaded and started.\n\n"
						+ "----------------------------------------");
			} else {
				throw new AgentCreationException(
					"The agent that was created was null.");
			}
		}
		System.out.println("loadAgent is done");
		return agent;
	}
	
	
	private void logLoadAgent(String msg) {
		log("loadAgent: " + msg);
	}
	
	/**
	 * Registers an agent on this LimoneServer.
	 */
	/*public void registerAgent(Agent agent) {
	 // add the new agent into the hashtable
	 agentTable.put(agent.getAgentID(), agent);
	 
	 // tell the agent to initially engage with all local agents
	 agent.setInitEng(getAgentProfiles());
	 
	 // refresh the beacon to be broadcasted
	 refreshBeacon();
	 }*/
	
	/**
	 * Returns the profiles of all agents currently running
	 * on this server.
	 *
	 * @return the profiles of all agents currently running
	 * on this server.
	 */
	private Profile[] getAgentProfiles() {
		Profile[] result = new Profile[agentTable.size()];
		Enumeration agentEnum = agentTable.elements();
		int i = 0;
		while (agentEnum.hasMoreElements()) {
			Agent currAgent = (Agent)agentEnum.nextElement();
			result[i] = currAgent.getProfile();
			i++;
		}
		return result;
	}
	
	/**
	 * Sets the time between each Beacon broadcast.
	 *
	 * @param period the period in milliseconds.
	 */
	/*public void setBeaconPeriod(long period) {
	 beaconPeriod = period;
	 if (booted) {
	 beaconServer.setPeriod(beaconPeriod);
	 beaconServer.startBroadcaster();
	 }
	 }*/
	
	/**
	 * Send a message to another agent.
	 *
	 * @param msg the message to be sent.
	 */
	public void sendMessage(Message msg) {
		log("Sending message to: " + msg.getDestHostID());
		
		if (msg.getDestHostID().equals(hostID))
			newMsg(msg);
		else
			msgSender.sendMessage(msg);
	}
	
	/**
	 * Implements the BeaconReceiver interface.  This is called whenever a
	 * beacon is received.  It distributes the beacon to each agent running
	 * on this server.
	 *
	 * @param be the BeaconEvent containing the beacon received.
	 */
	public void beaconReceived(BeaconEvent be) {
		
		//log("Received a beacon");
		
		Beacon beacon = be.getBeacon();
		
		if (beacon instanceof LimoneBeacon) {
			LimoneBeacon llb = (LimoneBeacon)beacon;
			
			Enumeration agentEnum = agentTable.elements();
			while(agentEnum.hasMoreElements()) {
				Agent a = (Agent)agentEnum.nextElement();
				
				//log("Notifying " + a.getAgentID() + " of new beacon");
				
				a.newBeacon(llb);
			}
		}
	}
	
	/**
	 * This is called by the Message Receiver whenever it receives a message.
	 * This method is synchronized with loadAgent to ensure that
	 * an agent has been fully created and registered with the LimoneServer
	 * before the server proceeds to process additonal messages.
	 *
	 * @param msg the message received by the Message Receiver.
	 */
	public void newMsg(Message msg) {
		
		if (msg.getType() == Message.LOAD_AGENT) {
			LoadAgentRequest lar = (LoadAgentRequest)msg;
			log("Loading agent " + lar.getAgentClassName());
			try {
				loadAgent(lar.getAgentClassName(), lar.getAgentName());
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
		} else {
			log("Received a new message for agent " + msg.getDestAgentID());
			AgentID aID = msg.getDestAgentID();
			Agent agent = (Agent)agentTable.get(aID);
			if (agent != null)
				agent.newMsg(msg);
			else
				log("Received a message for an agent not on this server!");
		}
	}
	
	/**
	 * Logs a debugging message.
	 *
	 * @param msg the message to log.
	 * @param level the level of the debug message.
	 */
	private void log(String msg) {
		StringBuffer label = new StringBuffer("LimoneServer");
		if (booted)
			label.append(" (").append(pv.port).append(")" );
		else
			label.append("[not booted]");
		
		DebugLogger.log(label.toString(), msg);
	}
	
	/**
	 * Returns true if this LimoneServer is running in debug mode.
	 */
	public boolean debug() {
		return pv.debug;
	}
	
	/**
	 * Returns a string representation of this class.
	 *
	 * @return a String representation of this class.
	 */
	public String toString() {
		String host = "";
		try{
			host = InetAddress.getLocalHost().toString();
		} catch (UnknownHostException e) {
			host = "Unkown Host";
		}
		
		StringBuffer sBuff = new StringBuffer(host);
		if (booted)
			sBuff.append(":").append(pv.port);
		else
			sBuff.append(":[not booted]");
		return sBuff.toString();
	}
}

