

package limone;

import java.io.*;
import eLights.*;

/**
 * A LocalTask is used to perform a remote operation.  It is inserted into the
 * queue within the AgentManager queue, which serializes the tasks and ensures
 * they are performed atomically.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
class LocalTask implements Serializable
{
	// requests by local agent
	public static final int INP = 0;
	public static final int INGP = 1;
	public static final int RDP = 2;
	public static final int RDGP = 3;
	public static final int OUT = 4;
	
	/**
	 * The type of the local task.  Possible values include
	 * LocalTask.INP, LocalTask.INGP, LocalTask.RDP, LocalTask.RDGP,
	 * and LocalTask.OUT.
	 */
	int type;
	
	/**
	 * True if this task was rejected by the destination agent.
	 */
	boolean rejected = false;
	
	/**
	 * True if this task timed out.
	 */
	boolean timedOut = false;
		
	/**
	 * The HostID of the host on which the destination agent is residing.
	 */
	HostID destHostID;
	
	/**
	 * The AgentID of the destination agent.
	 */
	AgentID destAgentID;
	
	/**
	 * The template used in performing a probing operation.
	 */
	ETemplate template;
	
	/**
	 * The resulting tuples in an probing group operation.
	 */
	ETuple[] tuples;
	
	/**
	 * The resulting tuples in a regular probing operation.
	 */
	ETuple tuple;
	
	/**
	 * true if the operation has completed.
	 */
	boolean hasResults = false;
	
	/**
	 * Creates a Task with the specified type.
	 *
	 * @param type the type of the task
	 * @param destHostID the destination host ID
	 * @param destAgentID the destination Agent ID
	 */
	public LocalTask(int type, HostID destHostID, AgentID destAgentID) {
		this.type = type;
		this.destHostID = destHostID;
		this.destAgentID = destAgentID;
	}
	
	/**
	 * Creates a Task with the specified type.
	 *
	 * @param type the type of the task
	 * @param destHostID the destination host ID
	 * @param destAgentID the destination Agent ID
	 * @param template the template used in the operation
	 */
	public LocalTask(int type, HostID destHostID, AgentID destAgentID,
					 ETemplate template) {
		this(type, destHostID, destAgentID);
		this.template = template;
	}
	
	/**
	 * Creates a Task with the specified type.
	 *
	 * @param type the type of the task
	 * @param destHostID the destination host ID
	 * @param destAgentID the destination Agent ID
	 */
	public LocalTask(int type, HostID destHostID, AgentID destAgentID,
					 ETuple tuple) {
		this(type, destHostID, destAgentID);
		this.tuple = tuple;
	}

	
	/**
	 * Sets this task to be done, releases any threads waiting for this
	 * task to be done.
	 */
	void taskDone() {
		hasResults = true;
		try {
			this.notifyAll();
		} catch(IllegalMonitorStateException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Does not allow the calling thread to return until this task is
	 * done.
	 */
	void holdUntilDone() {
		synchronized(this) {
			if (!hasResults) {
				try {
					wait();
				} catch(InterruptedException e) {
					e.printStackTrace();
					holdUntilDone();
				}
				if (!hasResults)
					holdUntilDone();
			}
		}
	}
	
	public String toString() {
		return "LocalTask:\n\tDestination AgentID: " + destAgentID
			+ "\n\tDestination HostID: " + destHostID;
	}
}
