package limone;

import java.util.*;
import java.io.*;
import eBeacon.*;
import limone.util.*;

/**
 * This contains all of the neighbors within communication range.
 * The acquaintance list is a subset of this list.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
class NeighborList implements Serializable
{
	/**
	 * Maps AgentIDs to Profiles.
	 */
    private Hashtable agentProfileTable = new Hashtable();
    
	/**
	 * Maps AgentIDs to long values representing the time
	 * the agent was last heard from.
	 */
	private Hashtable lastHeardTable = new Hashtable();
	
	/**
	 * A reference to the agent that this acquaintance list belongs to.
	 */
    private Agent a;
	
	/**
	 * Creates a Neighbor List.
	 */
	public NeighborList(Agent a) {
		this.a = a;
		
		// schedule a timer that checks for disconnection
		Timer timer = new Timer();
		timer.scheduleAtFixedRate(new NLTimer(), a.getDisengagementThreshold(),
								  a.getDisengagementThreshold());
	}
	
	/**
	 * Returns true  if the neighbor list contains a profile for the
	 * agent with the specified agent id.
	 */
	public boolean contains(AgentID aID) {
		
		//log("Checking to see if neighbor list contains " + aID);
		
		//printList();
		
		return agentProfileTable.containsKey(aID);
	}
	
	/**
	 * Adds a profile to the neighbor list.  This is called by the beacon
	 * manager.
	 */
	public void addProfile(Profile p) {
		//log("Adding the following profile: " + p);
		
		AgentID aID = ProfileParser.getAgentID(p);
		
		// update the time the profile was last received
		updateTime(aID);
		
		// put the agent into the neighbor list
		agentProfileTable.put(aID, p);
		
		// tell the acquaintance list that a new neighbor is present
		a.getAQL().profileAdded(p);
	}
	
	/**
	 * Removes the profile of the agent with the specified agent id from the
	 * neighbor list.
	 */
	public void removeProfile(AgentID aID) {
		
		// remove the agent from the lists
		Profile oldP = (Profile)agentProfileTable.remove(aID);
		lastHeardTable.remove(aID);
		
		log("Removed agent " + aID);
		
		if (oldP != null) {
			// tell the acquaintance list that a new neighbor is present
			a.getAQL().profileRemoved(oldP);
			a.getRL().removeRPs(aID);
		}
	}
	
	/**
	 * Removes a profile from the neighbor list.
	 * This is called by the beacon manager.
	 */
	public void removeProfile(Profile p) {
		removeProfile(ProfileParser.getAgentID(p));
	}
	
	/**
	 * Updatese the profile in the neighbor list.
	 * This is called by the beacon manager.
	 */
	public void updateProfile(Profile p) {
		
		log("Update profile called..");
		
		AgentID aID = ProfileParser.getAgentID(p);
		Profile oldP = (Profile)agentProfileTable.get(aID);
		
		if (oldP != null) {
			if (!oldP.equals(p)) {
				
				log("Updating profile for : " + aID);
				
				updateTime(aID);
				agentProfileTable.put(aID, p);
				a.getAQL().profileChanged(p);
				
			} else
				log("No need to update profile for " + aID);
		} else
			addProfile(p);
	}
	
	/**
	 * Gets the profile within this list of the agent with the specified
	 * agent ID.
	 */
	public Profile getProfile(AgentID aID) {
		return (Profile)agentProfileTable.get(aID);
	}
	
	/**
	 * Returns an enumeration of the agent profiles within this list.
	 */
	Enumeration elements() {
		return agentProfileTable.elements();
	}
	
	/**
	 * Updates the time that the last beacon was heard from the agent
	 * with the specified agent id.
	 */
	void updateTime(AgentID aID) {
		lastHeardTable.put(aID, new Long(System.currentTimeMillis()));
	}
	
	/**
	 * Returns the time that the last beacon was heard
	 * from the specified agent.  If the specified agent is not in the
	 * neighbor list, -1 is returned.
	 */
	long getLastHeard(AgentID aID) {
		//Long value = (Long)lastHeardTable.get(aID);
		//if(value == null) return -1;
		//else return value.longValue();
		return ((Long)lastHeardTable.get(aID)).longValue();
	}
	
	/**
	 * Prints the contents of the neighbor list.
	 */
	private void printList() {
		String result = "Contents of neighbor list:";
		
		Enumeration enum = agentProfileTable.keys();
		while(enum.hasMoreElements()) {
			AgentID aID = (AgentID)enum.nextElement();
			Profile p = (Profile)agentProfileTable.get(aID);
			result += "\n\t" + aID;//+ " - " + p;
		}
		log(result);
		
	}
	
	/**
	 * This checks for disconnection and removes profiles of agents whose
	 * beacons have not been received for an agent-defined period of time.
	 */
	private class NLTimer extends TimerTask {
		public NLTimer() {
		}
		
		public void run() {
			Enumeration keys = agentProfileTable.keys();
			
			// Go through each of the profiles within the neighbor list
			// and remove any whose last time in which the profile was received
			// is above the threshold defined within agent.
			
			while(keys.hasMoreElements()) {
				AgentID aID = (AgentID)keys.nextElement();
				//long lastHeard = getLastHeard(aID);
				//if(lastHeard < 0) continue;
				
				long time = System.currentTimeMillis() - getLastHeard(aID);
				if (time > a.getDisengagementThreshold())
					removeProfile(aID);
			}
			
		}
	}
	
	private void log(String msg) {
		DebugLogger.log("NeighborList", msg);
	}
}


