package limone;

import java.util.Vector;

/**
 * This is an implementation of a queue.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class Queue implements java.io.Serializable {

    private Vector taskQueue = new Vector();
    private Object lock = new Object();
    private boolean blocked = false;

    /**
     * Creates an empty queue.
     */
    public Queue() {}
	
	/**
	 * Returns true if the queue is empty.
	 */
	public synchronized boolean isEmpty() {
		return taskQueue.size() == 0;
	}
	
    /**
     * Adds an object to the end of this queue.
     */
    public synchronized void add(Object o) {
		taskQueue.addElement(o); wakeUpThreads();
	}

    /**
     * Removes the first item in the queue, or blocks if the queue is
     * empty or blocked.
     *
     * @return the object removed from the queue.
     */
    public Object remove() {
        Object result = null;

        while (result == null){
			if (taskQueue.size() == 0 || blocked)
				waitOnLock();
			
            synchronized(taskQueue) {
                if (taskQueue.size() > 0 && !blocked) {
                    result = taskQueue.elementAt(0);
					taskQueue.removeElementAt(0);
				}
            }
        }
		
        wakeUpThreads();
        return result;
    }

    /**
     * Blocks the queue.  This prevents the remove method from returning
     * anything but still allows the user to add objects into the queue.
     */
    public synchronized void block() { blocked = true; wakeUpThreads(); }

    /**
     * Unblocks the queue.  This enables objects to be removed from the queue.
     */
    public synchronized void unblock() { blocked = false; wakeUpThreads(); }

    /**
     * Blocks the calling thread until the queue is empty.
     */
    public void waitUntilEmpty() {
        while(taskQueue.size() != 0) {
            waitOnLock();
        }
    }

    /**
     * Halts the calling thread by making it wait on the lock.
     * If the thread is interrupted, re-hold on the lock.
     */
    private void waitOnLock() {
		
        synchronized(lock){
            try {
                lock.wait();
            } catch(InterruptedException ie) {
                ie.printStackTrace();
                waitOnLock();
            }
        }
    }

    /**
     * Calls notifyAll on the lock.
     */
    private void wakeUpThreads() {
		synchronized(lock) {
			lock.notifyAll();
		}
	}

    /**
     * Clears the queue.
     */
    public void clear() {taskQueue.removeAllElements(); wakeUpThreads();}
	
	private void log(String msg) {
		System.out.println("Queue: " + msg);
		System.out.flush();
	}
}
