package limone;

/**
 * A Reaction is used by an Agent to tell another Agent that it
 * wants to know about a tuple matching a particular template.
 * It can be ONCE or ONCE-PER-TUPLE. A ONCE Reaction only fires only once,
 * whereas a ONCE-PER-TUPLE reaction fires once for each matching tuple.
 * A ONCE-PER-TUPLE Reaction does not maintain any history.  Thus when an
 * Agent disengages and then re-engages, a ONCE-PER-TUPLE reaction
 * will re-react to all of the matching tuples in the Agent's Tuple Space.
 *
 * <p>A Reaction consists of two components:  the Reactive Pattern and the
 * call-back function.  The reactive patterns determines where the reaction
 * is propagated and what tuples it is sensitive to.  The call-back function
 * is code that is executed whenever a tuple it is sensitive to is found.
 *
 * @see ReactivePattern
 * @see ReactionListener
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class Reaction implements java.io.Serializable{
	
	/**
	 * Defines a reaction to be of type ONCE, meaning it will automatically
	 * be deregistered once fired.
	 */
    public static final int ONCE = 0;
	
	/**
	 * Defines a reaction to be of type ONCE_PER_TUPLE, meaning it will
	 * remain registered even after firing.
	 */
    public static final int ONCE_PER_TUPLE = 1;
    
    /**
     * The ReactivePattern within this Reaction.
     */
    ReactivePattern rp;
    
    /**
     * The ReactionListener within this Reaction.
     */
    ReactionListener rl;
	
	/**
	 * This reaction's Reaction ID.
	 */
	ReactionID rID;
	
    
    /**
     * Creates a Reaction with the specified reactive pattern
     * and reaction listener.
     *
     * @param rp the reactive pattern of this reaction.
     * @param rl the reaction listener of the reaction.
     */
    public Reaction(ReactivePattern rp, ReactionListener rl) {
        this.rp = rp;
        this.rl = rl;
    }
	
	/**
	 * Sets the reaction ID of this reaction.  This ID is unique for each
	 * reaction created by this agent.  It is not globally unique.
	 */
	public void setReactionID(ReactionID rID) {
		this.rID = rID;
		rp.rID = rID;
	}
    
    /**
     * Returns the callback function within this reaction.
     *
     * @return the callback function within this reaction.
     */
    public ReactionListener getCF() {
        return rl;
    }
    
    /**
     * Determines whether another object is equivalent to this object.
     * Another object is equivalent to this object when the other
     * object if of type Reaction and has the same ReactionID and
     * ReactivePattern as this one.
     *
     * @return true if the other object is equivalent to this object.
     */
    public boolean equals(Object o) {
        if (o instanceof Reaction) {
            Reaction rxn = (Reaction)o;
            return rxn.rp.rID.equals(rp.rID) &&
				rxn.rp.getAgentID().equals(rp.getAgentID());
        }
        return false;
    }
    
    /**
     * Returns a String representation of this class.
     *
     * @return a String representation of this class.
     */
    public String toString() {
        return rp.toString();
    }
}
