package limone;

import eLights.*;
import eBeacon.*;
import java.util.*;
import java.io.*;
import limone.io.*;

/**
 * The Reaction List is a list of Reactive Patterns that apply to the
 * local Tuple Space.  These Reactive Patterns may be from the local
 * Agent as well as remote Agents.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
class ReactionList implements Serializable {
	/**
	 * A list of Reactive Patterns.
	 */
    private Vector rList = new Vector();
    
	/**
	 * The agent that this list belongs to.
	 */
	private Agent a;
	
	/**
	 * A reference to the LimoneServer.
	 */
	private LimoneServer ls;
	
    /**
	 * Creates a ReactionList.
	 */
    public ReactionList(Agent a) {
		this.a = a;
		this.ls = LimoneServer.getServer();
	}
    
	/**
	 * Returns true if this reaction list contains the specified reactive
	 * pattern.
	 */
	boolean contains(ReactivePattern rp) {
		for (int i = 0; i < rList.size(); i++) {
			ReactivePattern rPat = (ReactivePattern)rList.elementAt(i);
			if (rPat.equals(rp))
				return true;
		}
		return false;
	}
	
    /**
	 * Adds a reactive pattern into the list if the reactive pattern is not
	 * already in the list, as determined by ReactionID and AgentID.
	 *
	 * @return true if the reactive pattern was added to the list.
	 */
    boolean addRP(ReactivePattern rp) {
		if (!contains(rp)) {
			
			rList.addElement(rp);
			log("Added the reactive pattern to the reaction list.");
												
			
			log("Checking to see if any tuples currently in the tuple space"
					+ " should fire the reaction.");
			ETuple[] tuples = a.getTupleSpace().rdgp(rp.getTemplate());
			
			if (tuples != null && tuples.length > 0) {
				
				log(tuples.length + " matching tuples");
								
				boolean go = true;
				for (int i = 0; i < tuples.length && go; i++) {
					go = fireReaction(rp, tuples[i]);
					log("4");
				}
				
			} else
				log("No matching tuples");
						
			log("done addRP");
								
			return true;
		} else
			return false;
    }
	
	/**
	 * For each tuple in the specified array, check to see if any
	 * of the reactive patterns within this list is
	 * sensitive to the specified tuple.  If so, fires the reaction.  If
	 * the reaction is type ONCE, then it is removed from the list.
	 * Called by the tuple space when a new tuple was inserted into it.
	 *
	 * @param t the newly added tuple.
	 */
	void newTuples(ETuple[] t) {
		log("newTuples called, number of new tuples = " + t.length);
		for (int i = 0; i < t.length; i++) {
			newTuple(t[i]);
		}
	}
	
	/**
	 * Check to see if any of the reactive patterns within this list is
	 * sensitive to the specified tuple.  If so, fires the reaction.  If
	 * the reaction is type ONCE, then it is removed from the list.
	 * Called by the tuple space when a new tuple was inserted into it.
	 *
	 * @param t the newly added tuple.
	 */
	void newTuple(ETuple t) {
		
		log("newTuple called, rList.size = " + rList.size());
		
		// for each reactive pattern within this list, if the
		// newly added tuple matches the reactive pattern's template,
		// fire the reaction
		for (int i = 0; i < rList.size(); i++) {
			ReactivePattern rp = (ReactivePattern)rList.elementAt(i);
			ETemplate temp = rp.getTemplate();
			
			
			if (temp.matches(t)){
				log("The template matches the tuple!!");
				fireReaction(rp, t);
			} else
				log("The template did not match the tuple");
			
		}
	}
	
	/**
	 * Sends a message to the agent that registered the reactive pattern
	 * telling it that the reaction was fired by the tuple.
	 *
	 * @return true if the reactive pattern was a ONCE reaction and thus removed
	 * from the reaction list.
	 */
	private boolean fireReaction(ReactivePattern rp, ETuple t) {
		ReactionResults rr = new ReactionResults(
			rp.getHostID(), rp.getAgentID(), a.getAgentID(), rp.rID, t);
		
		log("Sending the reaction results message to "
				+ rp.getAgentID());
		ls.sendMessage(rr);
		
		// the reaction that was just fired is a ONCE reaction,
		// remove it from this list
		if(rp.type == Reaction.ONCE) {
			removeRP(rp);
			return false;
		} else
			return true;
		
	}
    
    /**
	 * Removes a ReactivePattern from the list.
	 *
	 * @param rp the ReactivePattern to be removed.
	 */
    void removeRP(ReactivePattern rp) {
		rList.removeElement(rp);
    }
    
    /**
	 * Removes all ReactivePatterns created by the Agent represented by
	 * AgentID aID from the list.
	 *
	 * @param aID the AgentID of the Agent that registered the
	 * reactions to be removed.
	 */
    void removeRPs(AgentID aID) {
		for (int i = 0; i < rList.size(); i++) {
			ReactivePattern rPat = (ReactivePattern)rList.elementAt(i);
			if (rPat.getAgentID().equals(aID)) {
				rList.removeElementAt(i);
				i--;
			}
		}
    }
	
	
    /**
	 * Returns a String representation of this class.
	 *
	 * @return a String representation of this class.
	 */
    public String toString() {
		String result = "ReactionList:";
		
		for (int i = 0; i < rList.size(); i++) {
			result += "\n\t" + rList.elementAt(i);
		}
		return result;
    }
	
	void log(String msg) {
		DebugLogger.log("ReactionList", msg);
	}
}


