package limone;

import java.io.*;
import eBeacon.*;
import limone.util.*;
import limone.io.*;

/**
 * The reaction manager is responsible for propagating and un-propagating
 * reactions, and executing reaction call back functions when they fire.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
class ReactionManager implements Runnable, Serializable
{
	private static final int ADD_RXN = 0;
	private static final int REMOVE_RXN = 1;
	private static final int ADD_AGENT = 2;
	private static final int REMOVE_AGENT = 3;
	
	private Queue taskQueue = new Queue();
	
	private Agent a;
	
	/**
	 * This queue holds pending reactions that have fired but
	 * are awaiting the callback function to be executed.
	 */
	private Queue rxnQueue = new Queue();
	
	/**
	 * This is the thread that executes the reaction call back function.
	 */
	private RxnThread rt;
	
	/**
	 * Creates a reaction manager for the specified agent.
	 */
	public ReactionManager(Agent a) {
		this.a = a;
		rt = new RxnThread();
		new Thread(this).start();
	}
	
	/**
	 * This is called by the agent when it recieves the results of
	 * a reaction.
	 */
	void processRxnResults(ReactionResults rr) {
		log("processRxnResults(...) called");
		rxnQueue.add(rr);
	}
	
	/**
	 * This is called by the reaction registry when a reaction is added to it.
	 *
	 * @param r the reaction that was added.
	 */
	void addReaction(Reaction r) {
		log("addReaction(...) called");
		taskQueue.add(new RMTask(ADD_RXN, r));
	}
	
	/**
	 * This is called by the reaction registry when a reaction is removed
	 * from it.
	 *
	 * @param r the reaction that was removed.
	 */
	void removeReaction(Reaction r) {
		taskQueue.add(new RMTask(REMOVE_RXN, r));
	}
	
	/**
	 * This is called by the aquaintance list when an agent is added to it.
	 *
	 * @param p the profile of the newly added agent.
	 */
	void agentAdded(Profile p) {
		taskQueue.add(new RMTask(ADD_AGENT, p));
	}
	
	/**
	 * This is called by the acquaintance list when an agent is removed from it.
	 *
	 * @param p the profile of the agent that was removed.
	 */
	void agentRemoved(Profile p) {
		taskQueue.add(new RMTask(REMOVE_AGENT, p));
	}
	
	public void run() {
		
		//System.out.println("REACTION MANAGER RUN METHOD CALLED");
		
		LimoneServer ls = LimoneServer.getServer();
		
		while(true) {
			RMTask nxtTask = (RMTask)taskQueue.remove();
			switch(nxtTask.type) {
				
				case ADD_RXN:
					
					// get the profiles of the agents that the reaction should
					// be propagated to
					log("Rxn added, getting matching profiles.");
					Profile[] p1 = a.getAQL()
						.getAgentProfiles(nxtTask.r.rp.pss);
					
					log("There were " + p1.length + " matches");
					
					// send a message to each of these agents asking
					// them to register the reaction
					for (int i = 0; i < p1.length; i++) {
						Request req = new Request(
							Message.REGISTER_REACTIVE_PATTERN,
							ProfileParser.getAgentID(p1[i]),
							ProfileParser.getHostID(p1[i]),
							a,
							nxtTask.r.rp);
						ls.sendMessage(req);
					}
					break;
					
				case REMOVE_RXN:
					
					// get the profiles of the agents that the reaction was
					// propagated to
					Profile[] p2 = a.getAQL()
						.getAgentProfiles(nxtTask.r.rp.pss);
					
					// send a message to each of these agents asking
					// them to deregister the reaction
					for (int i = 0; i < p2.length; i++) {
						Request req = new Request(
							Message.DEREGISTER_REACTIVE_PATTERN,
							ProfileParser.getAgentID(p2[i]),
							ProfileParser.getHostID(p2[i]),
							a,
							nxtTask.r.rp);
						ls.sendMessage(req);
					}
					
					
					break;
					
				case ADD_AGENT:
					log("Agent added, getting matching reactions.");
					
					// get the reactions that should be propagated
					// to the newly added agent
					Reaction[] rxns1 = a.getRR().get(nxtTask.p);
					
					log("There were " + rxns1.length + " matches");
					
					// send a message to the newly added agent asking
					// it to register the reaction
					for (int i = 0; i < rxns1.length; i++) {
						Request req = new Request(
							Message.REGISTER_REACTIVE_PATTERN,
							ProfileParser.getAgentID(nxtTask.p),
							ProfileParser.getHostID(nxtTask.p),
							a,
							rxns1[i].rp);
						ls.sendMessage(req);
					}
					
					break;
					
				case REMOVE_AGENT:
					
					// get the reactions that should be un-propagated
					// to the newly added agent
					Reaction[] rxns2 = a.getRR().get(nxtTask.p);
					
					// send a message to the newly added agent asking
					// it to deregister the reaction
					for (int i = 0; i < rxns2.length; i++) {
						Request req = new Request(
							Message.DEREGISTER_REACTIVE_PATTERN,
							ProfileParser.getAgentID(nxtTask.p),
							ProfileParser.getHostID(nxtTask.p),
							a,
							rxns2[i].rp);
						ls.sendMessage(req);
					}
					
					break;
			}
		}
	}
	
	/**
	 * Used to tell the reaction manager thread to do a certain operation.
	 */
	private class RMTask implements Serializable {
		int type;
		
		/**
		 * The profile of the agent that was added or removed from the
		 * acquaintance list
		 */
		Profile p;
		
		/**
		 * The reaction that was added or removed from the reaction registry.
		 */
		Reaction r;
		
		public RMTask(int type, Profile p) {
			this.type = type;
			this.p = p;
		}
		
		public RMTask(int type, Reaction r) {
			this.type = type;
			this.r = r;
		}
	}
	
	/**
	 * This is the reaction thread that executes the reaction callback
	 * functions when they fire.
	 */
	private class RxnThread extends Thread {
		public RxnThread() {
			start();
		}
		
		public void run() {
			while(true) {
				log("Waiting for next reaction results");
				ReactionResults rr = (ReactionResults)rxnQueue.remove();
				
				ReactionID rID = rr.getRxnID();
				Reaction r = a.getRR().get(rID);
				
				log("Got the next reaction results, rID = " + rID);
				
				if (r != null) {
					
					// create the reaction event
					ReactionEvent re = new ReactionEvent(
						r.rp,
						rr.getTuple(),
						rr.getSourceAgentID());
					
					// execute the callback function
					r.getCF().reactsTo(re);
					
					if (r.rp.type == r.ONCE) {
						a.getRR().deregisterReaction(r);
					}
				} else
					log("The reaction is no longer in the reaction registry.");
				
			}
		}
		
		void log(String msg) {
			DebugLogger.log("ReactionManager: RxnThread", msg);
		}
	}
	
	void log(String msg) {
		DebugLogger.log("ReactionManager", msg);
	}
}


