package limone;

import java.io.*;
import java.util.*;
import eBeacon.*;

/**
 * ReactionRegistry.java
 *
 * A Reaction Registry is a list of all Reactions created and registered
 * by the Agent.  It is another object that is inherited when extending
 * Agent.  By registering a reaction in the reaction registry, Limone
 * automatically propagates the reaction to remote agents and handles
 * the execution of the reaction's callback function when their fire.
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public class ReactionRegistry implements Serializable
{
	
	/**
	 * The internal data structure of this ReactionRegistry.
	 */
    Vector rr = new Vector();
    
	/**
	 * The agent using this reaction registry.
	 */
	Agent a;
	
    /**
	 * Creates a ReactionRegistry.
	 */
    public ReactionRegistry(Agent a) {
		this.a = a;
    }
    
    /**
	 * Adds a Reaction to this ReactionRegistry.
	 *
	 * @param rxn the Reaction to add.
	 * @param the ReactionID of the reaction just registered
	 */
    public ReactionID registerReaction(Reaction rxn) {
		log("registerReaction(...) called");
		
		// set the agent location of the reaction
		rxn.rp.aLoc = new AgentLocation(a.getAgentID(), a.getHostID());
		
		// set the reaction's reaction ID
		rxn.rp.rID = ReactionID.generateNew();
		
		rr.addElement(rxn);
		
		log("passing the reaction to the reaction manager");
		a.getRxnMgr().addReaction(rxn);
		return rxn.rp.getRID();
    }
    
    /**
	 * Removes a Reaction to this ReactionRegistry.
	 *
	 * @param rxn the Reaction to remove.
	 * @return the reaction that was removed, or null if the reaction did
	 * not exist.
	 */
    public Reaction deregisterReaction(Reaction rxn) {
		if (rr.removeElement(rxn)) {
			a.getRxnMgr().removeReaction(rxn);
			return rxn;
		} else
			return null;
    }
    
    /**
	 * Removes a Reaction to this ReactionRegistry.
	 *
	 * @param rID the ReactionID of the Reaction to be removed.
	 * @return the reaction that was removed,
	 * or null if the reaction did not exist.
	 */
    public Reaction deregisterReaction(ReactionID rID) {
		Reaction rxn = get(rID);
		if (rxn != null)
			deregisterReaction(rxn);
		return rxn;
    }
    
    /**
	 * Retrieves all Reactions that apply to a particular profile.
	 *
	 * @param profile the Profile that all of the reactions
	 * we want to obtain apply to.
	 * @return an array of all of the Reactions apply to AgentLocation aLoc.
	 */
    Reaction[] get(Profile profile) {
		Vector results = new Vector();
		for (int i = 0; i < rr.size(); i++) {
			Reaction curr = (Reaction)rr.elementAt(i);
			if (curr.rp.appliesTo(profile))
				results.addElement(curr);
		}
		Reaction[] result = new Reaction[results.size()];
		for (int i = 0; i < results.size(); i++) {
			result[i] = (Reaction)results.elementAt(i);
		}
		return result;
    }
    
    /**
	 * Retrieves a reaction with ReactionID rID.
	 *
	 * @param rID the ReactionID of the Reaction to
	 * be retrieved from this registry.
	 * @return The reaction retrieved, or null if no Reaction
	 * had a ReactionID of rID in the registry.
	 */
    Reaction get(ReactionID rID) {
		for (int i = 0; i < rr.size(); i++) {
			Reaction curr = (Reaction)rr.elementAt(i);
			if (curr.rp.rID.equals(rID))
				return curr;
		}
		return null;
    }
    
    /**
	 * Returns a String representation of this class.
	 *
	 * @return a String representation of this class.
	 */
    public String toString() {
		return "ReactionRegistry";
    }
	
	void log(String msg) {
		DebugLogger.log("ReactionRegistry", msg);
	}
}

