package limone;

import eLights.*;
import eBeacon.*;

/**
 * A Reactive Pattern is contained within a Reaction.
 * It is used to determine where a reaction should propagate to and what
 * tuples it is sensitive to.
 *
 * @see Reaction
 *
 * @author Chien-Liang Fok
 * @version 03/17/2003
 */
public class ReactivePattern implements java.io.Serializable{
	/**
	 * The current location of the agent that created
	 * this ReactivePattern.
	 */
	AgentLocation aLoc;
	
	/**
	 * Specifies where this ReactivePattern should be registered.
	 */
	ProfileSelector[] pss;
	
	/**
	 * Distinguishes this ReactivePattern from all others created
	 * by the same agent.
	 */
	ReactionID rID;
	
	/**
	 * Can be Reaction.ONCE or Reaction.ONCE_PER_TUPLE
	 */
	int type;
	
	/**
	 * The template with the two system fields added to the beginning of it.
	 */
	ETemplate template;
	
	/**
	 * Creates a reactive pattern.
	 *
	 * @param ps specifies where the reactive pattern should be registered.
	 * @param type either Reaction.ONCE or Reaction.ONCE_PER_TUPLE
	 * @param template specifies which tuples this Reactive Pattern
	 * is sensitive to.
	 */
	public ReactivePattern(ProfileSelector ps, int type, ETemplate template) {
		this(new ProfileSelector[] {ps}, type, template);
	}
	
	/**
	 * Creates a reactive pattern.
	 *
	 * @param pss specifies where the reactive pattern should be registered.
	 * @param type either ONCE or ONCE_PER_TUPLE
	 * @param template specifies which tuples this Reactive Pattern
	 * is sensitive to.
	 */
	public ReactivePattern(ProfileSelector[] pss, int type,
						   ETemplate template) {
		this.pss = pss;
		this.type = type;
		this.template = template;
	}
	
	/**
	 * Returns this reactive pattern's reaction ID.
	 */
	public ReactionID getRID() {
		return rID;
	}
	
    /**
	 * Returns true if this reactive pattern is sensitive to the
	 * specified tuple.
	 *
	 * @param tuple the tuple to compare with.
	 * @return true if this ReactivePattern applies to the specified tuple.
	 */
    public boolean appliesTo(ETuple tuple) {
		return template.matches(tuple);
    }
	
    /**
	 * Determines whether this reactive pattern should be
	 * propagated to the agent with the specified profile.
	 *
	 * @param profile the profile of the agent.
	 * @return true if this reactive pattern should be registered on the agent.
	 */
    public boolean appliesTo(Profile profile) {
		boolean result = false;
		for (int i = 0; i < pss.length && !result; i++) {
			if (pss[i].matches(profile))
				result = true;
		}
		return result;
    }
    
	/**
	 * Returns the AgentID of the agent that created this reactive pattern.
	 * @return the AgentID of the agent that created this reactive pattern.
	 */
	public AgentID getAgentID() {
		return aLoc.agentID();
	}
	
    /**
	 * Returns the HostID of the agent that created this reactive pattern.
	 *
	 * @return the HostID of the agent that created this reactive pattern.
	 */
    public HostID getHostID() {
		return aLoc.hostID();
    }
    
    /**
	 * Returns the profiles that determine where this reactive pattern should
	 * be registered.
	 *
	 * @return the profiles that determine where this reactive pattern should
	 * be registered.
	 */
    public ProfileSelector[] getProfileSelectors() {
		return pss;
    }
	
	
	/**
	 * Returns the Template contained within this reactive pattern.
	 * The template determines which tuples this reactive pattern is
	 * sensitive to.  When such a tuple appears in the tuple space,
	 * the reaction should fire.
	 *
	 * @return the Template contained within this reactive pattern.
	 */
	public ETemplate getTemplate() {
		return template;
	}
	
	/**
	 * Another object is equal to this one when the other object
	 * if of type ReactivePattern and they have the same reaction ID.
	 *
	 * @param the object to be compared to this one.
	 * @return true if the object is equal to this one.
	 */
	public boolean equals(Object o) {
		if (o instanceof ReactivePattern) {
			ReactivePattern rp = (ReactivePattern)o;
			return getAgentID().equals(rp.getAgentID()) &&
				rp.rID.equals(rID);
		}
		return false;
	}
	
    /**
	 * Sets the agent location of the reactive pattern.  This method can only
	 * be called once.  It is called by the agent that registered the reactive pattern.
	 * It is used by the Agent that created the reaction to determine on which agent
	 * the reactive pattern fired.
	 *
	 * <p> This method is called by the underlying LimeLite system when the reaction
	 * containing this reactive pattern is registered on a tuple space.
	 *
	 * @param aLoc the AgentLocation of the agent that registered this reactive pattern.
	 * @throws ReactionException when this method is called more than once.
	 */
    public void setAgentLocation(AgentLocation aLoc) throws ReactionException {
		if (this.aLoc == null)
			this.aLoc = aLoc;
		else
			throw new ReactionException(
				"The Reactive Pattern already had an Agent Location.");
    }
	
	/**
	 * Returns a String representation of this class.
	 *
	 * @return a String representation of this class.
	 */
	public String toString() {
		String result = "[";
		if (type == Reaction.ONCE)
			result += " ONCE";
		else
			result += " ONCE_PER_TUPLE";
		result += " " + rID + ", template = " + template + "]";
		return result;
	}
}

