package limone;

import java.io.*;
import java.util.*;
import limone.io.*;

/**
 * The RemoteOpManager is responsible for manageming the operation requests
 * being sent to remote agents.  It ensures that there is at most one
 * outstanding request sent to each remote agent and that the remote agent is
 * within the acquaintance list.  It also times out the requests when a reply
 * was not received within a certain period of time as defined by the
 * agent.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
class RemoteOpManager implements Serializable
{
	/**
	 * The local Limone Server.
	 */
	LimoneServer ls;
	
	/**
	 * The agent that this manager belongs to.
	 */
	Agent a;
	
	/**
	 * The opTable maps an AgentID to an RemoteOpMgrThread.
	 */
	Hashtable opTable = new Hashtable();
	
	/**
	 * Creates a RemoteOpManager for the specified agent.
	 */
	public RemoteOpManager(Agent a) {
		this.a = a;
		ls = LimoneServer.getServer();
	}
	
	/**
	 * process a reply message.
	 */
	public void processReply(Reply reply) {
		AgentID aID = reply.getSourceAgentID();
		RemoteOpMgrThread romt = (RemoteOpMgrThread)opTable.get(aID);
		
		if (romt != null)
			romt.processReply(reply);
		
	}
	
	/**
	 * Adds a task that should be processed locally.
	 */
	public void addTask(LocalTask task) {
		AgentID aID = task.destAgentID;
		synchronized(opTable) {
			if (opTable.containsKey(aID)) {
				RemoteOpMgrThread romt = (RemoteOpMgrThread)opTable.get(aID);
				romt.addTask(task);
			} else {
				RemoteOpMgrThread romt = new RemoteOpMgrThread(aID, task);
				opTable.put(aID, romt);
				new Thread(romt).start(); // start the thread
			}
		}
	}
	
	/**
	 * Handles the requests made to a particular agent.
	 */
	private class RemoteOpMgrThread implements Runnable, Serializable {
		
		/**
		 * The agentID of the agent that this remoteOpMgrThread communicates
		 * with.
		 */
		AgentID aID;
		
		/**
		 * Enqueues the tasks that need to be performed on the agent this
		 * thread is responsible for.
		 */
		Queue taskQueue = new Queue();
		
		/**
		 * The lock that this thread blocks on while waiting for a reply.
		 */
		Object lock = new Object();
		
		/**
		 * This is the timer that times out an operation should the reply not
		 * be received in time.
		 */
		Timer timeout;
		
		/**
		 * This is the reply that we received from the remote agent,
		 * or we contocted due to timing out.
		 */
		Reply reply;
		
		/**
		 * This is the enumerated value of the results that we are expecting.
		 */
		int expectedID;
		
		/**
		 * This is the current task being performed.
		 */
		LocalTask currTask;
		
		/**
		 * Creates a RemoteOpMgrThread.  This thread handles the sending of
		 * requests and receiving of replies to and from a particular
		 * remote agent.
		 *
		 * aID the
		 */
		public RemoteOpMgrThread(AgentID aID, LocalTask initialTask) {
			this.aID = aID;
			taskQueue.add(initialTask);
		}
		
		/**
		 * Processes a reply received from a remote agent.
		 */
		public void processReply(Reply reply) {
			
			log("processReply: Processing reply");
			
			// If the reply does not have the expected ID, it must
			// be one that arrived too late.  Simply discard it.
			if (expectedID == reply.getID()) {
				
				log("processReply: Synchronizing on lock");
				
				synchronized(lock) {
					
					log("processReply: Synchronized on lock");
					
					if (this.reply == null) {
																		
						this.reply = reply;
						
						try {
							log("processReply: Notifying the lock");
							lock.notifyAll();
						} catch(Exception e) {
							e.printStackTrace();
						}
						
					} else
						log("processReply: Already received reply");
				}
			} else
				log("processReply: Not the expected ID\n\t"
						+ "Expected: " + expectedID
						+ "\n\tReceived: " + reply.getID());
			
		}
		
		/**
		 * Adds a task that needs to be performed.
		 */
		public void addTask(LocalTask task) {
			taskQueue.add(task);
		}
		
		/**
		 * Wait on the lock until a reply is received.
		 */
		private  void waitOnLock() {
			try {
				synchronized(lock) {
					if (reply == null)
						lock.wait();
				}
			} catch(InterruptedException e) {
				e.printStackTrace();
				if (reply == null)
					waitOnLock();
			}
		}
		
		/**
		 * A thread executes the loop defined within this method
		 * that processes the local requests.
		 */
		public void run() {
			
			while(!taskQueue.isEmpty()) {
				
				reply = null;
				
				currTask = (LocalTask)taskQueue.remove();
				
				// create the request
				Request currRequest = null;
				
				switch(currTask.type) {
					case LocalTask.OUT:
						log("Created a new OUT request");
						currRequest = new Request(
							currTask.destAgentID, currTask.destHostID, a,
							currTask.tuple);
						break;
					case LocalTask.INP:
						currRequest = new Request(
							Message.INP, currTask.destAgentID,
							currTask.destHostID, a, currTask.template);
						break;
					case LocalTask.INGP:
						currRequest = new Request(
							Message.INGP, currTask.destAgentID,
							currTask.destHostID, a, currTask.template);
						break;
					case LocalTask.RDP:
						currRequest = new Request(
							Message.RDP, currTask.destAgentID,
							currTask.destHostID, a, currTask.template);
						break;
					case LocalTask.RDGP:
						currRequest = new Request(
							Message.RDGP, currTask.destAgentID,
							currTask.destHostID, a, currTask.template);
				}
				
				// set the expected ID
				expectedID = currRequest.getID();
				
				// send the request
				ls.sendMessage(currRequest);
				
				log("Sent the request, ID = " + currRequest.getID());
				
				// set the timer
				timeout = new Timer();
				long delay = a.getOperationTimeout();
				timeout.schedule(new TimeOuter(this, currRequest),
								 delay);
				
				log("Timer has been set, delay = "
						+ delay + "ms, waiting on the lock.");
				
				// block
				waitOnLock();
				
				log("Reply received");
				
				// cancel the timer
				timeout.cancel();
				timeout = null;
				
				log("Timer cancelled");
				
				// check to see if the request timed out or was rejected
				if (reply.timedOut()) {
					
					log("The request timed out");
					currTask.timedOut = true;
					
				} else if (reply.isRejected()) {
					
					log("The request was rejected");
					currTask.rejected = true;
					
				}
					
					// if the current task is a probing operation,
					// retrieve the resulting tuple
				else if (currTask.type == LocalTask.INP ||
						 currTask.type == LocalTask.RDP) {
					currTask.tuple = reply.getTuple();
				}
					
					// if the current task is a probing group operation,
					// retrieve the resulting tuples
				else if (currTask.type == LocalTask.INGP ||
						 currTask.type == LocalTask.RDGP) {
					currTask.tuples = reply.getTuples();
				}
				
				// Release the thread that was waiting on the current
				// task to complete
				synchronized(currTask) {
					try {
						
						log("Notifying currTask");
						currTask.taskDone();
						currTask.notifyAll();
					} catch(IllegalMonitorStateException e) {
						e.printStackTrace();
					}
				}
				
				log("Done with currTask");
				this.currTask = null;
				this.reply = null;
				
			} // while...
			
			// remove this thread from the hashtable
			synchronized(opTable) {
				opTable.remove(aID);
			}
		}
		
		void log(String msg) {
			DebugLogger.log("RemoteOpManager: RemoteOpMgrThread:", msg);
		}
	} // class RemoteOpMgrThread
	
	private class TimeOuter extends TimerTask implements Serializable {
		RemoteOpMgrThread romt;
		Request request;
		
		public TimeOuter(RemoteOpMgrThread romt, Request request) {
			this.romt = romt;
			this.request = request;
		}
		
		public void run() {
			
			log("TimeOuter: running run() method");
			
			Reply timeOutReply = new Reply(request.getType(), request.getID(),
										   request.getDestHostID(),
										   request.getDestAgentID(),
										   null, false);
			timeOutReply.setTimedOut();
			romt.processReply(timeOutReply);
		}
	}
	
	private void log(String msg) {
		DebugLogger.log("RemoteOpManager", msg);
	}
}
