package limone.io;

import java.io.*;
import eLights.*;
import limone.*;

/**
 * Reply.java
 *
 * This is sent from an agent in response to a request by another agent.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class Reply implements Message, Serializable
{
	/**
	 * The type of reply.
	 */
	int type;
	
	/**
	 * Indicates that the request was rejected by the receipient.
	 */
	boolean isRejected;
	
	/**
	 * Indicates that that request timed out.
	 */
	boolean timedOut;
	
	/**
	 * A unique integer that indentifies this request.  It is used to correlate
	 * the request with the response.
	 */
	int id;
	
	/**
	 * The AgentID of the agent sending this reply.
	 */
	AgentID sAgentID;
	
	/**
	 * The AgentID of the agent that this reply is destined to.
	 */
	AgentID dAgentID;
	
	/**
	 * The HostID of the LimoneServer on which the agent this reply is destined
	 * to is residing on.
	 */
	HostID dHostID;
	
	/**
	 * The resulting tuples from a probing group operation.
	 */
	ETuple[] tuples;
	
	/**
	 * The resulting tuples from a probig operation.
	 */
	ETuple tuple;
	
	
	/**
	 * The default constructor does not do anything.
	 */
	public Reply() {
	}
	
	/**
	 * Creates a reply.
	 *
	 * @param type the type of the operation being replied to. Possible values
	 * include Message.OUT, Message.REGISTER_REACTIVE_PATTERN and
	 * Message.DEREGISTER_REACTIVE_PATTERN.
	 * @param id the id of the request that this reply is for
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param sAgentID the AgentID of the agent sending this reply
	 */
	public Reply(int type, int id, HostID destHostID, AgentID destAgentID,
				 AgentID sAgentID) {
		this.type = type;
		this.id = id;
		this.dHostID = destHostID;
		this.dAgentID = destAgentID;
		this.sAgentID = sAgentID;
	}
	
	/**
	 * Creates a reply that indicates a request has been rejected.
	 *
	 * @param type the type of the operation being replied to.
	 * @param id the id of the request that this reply is for
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param sAgentID the AgentID of the agent sending this reply
	 * @param isRejected true if the request was rejected
	 */
	public Reply(int type, int id, HostID destHostID, AgentID destAgentID,
				 AgentID sAgentID, boolean isRejected) {
		this(type, id, destHostID, destAgentID, sAgentID);
		this.isRejected = isRejected;
	}
	
	/**
	 * Creates a reply that indicates a request has been rejected.
	 *
	 * @param type the type of the operation being replied to.
	 * @param id the id of the request that this reply is for
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param sAgentID the AgentID of the agent sending this reply
	 * @param tuples the tuples that resulted from a probing group operation.
	 */
	public Reply(int type, int id, HostID destHostID, AgentID destAgentID,
				 AgentID sAgentID, ETuple[] tuples) {
		this(type, id, destHostID, destAgentID, sAgentID);
		this.tuples = tuples;
	}
	
	/**
	 * Creates a reply containing a tuple.  This can be the results of an
	 * INp or RDp operation.
	 *
	 * @param type the type of the operation being replied to.
	 * @param id the id of the request that this reply is for
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param sAgentID the AgentID of the agent sending this reply
	 * @param tuple the tuple that resulted from a probing operation.
	 */
	public Reply(int type, int id, HostID destHostID, AgentID destAgentID,
				 AgentID sAgentID, ETuple tuple) {
		this(type, id, destHostID, destAgentID, sAgentID);
		this.tuple = tuple;
	}
	
	/**
	 * Returns the enumerated id of the request that this reply is replying
	 * to.
	 */
	public int getID() {
		return id;
	}
	
	/**
	 * Returns the type of the message.  Type possible
	 * types are defined within Message.
	 *
	 * @return the type of the message.  Type possible
	 * types are defined within Message.
	 */
    public int getType() {
		return type;
	}
    
    /**
	 * Returns the destination hostID of the host that this message
	 * should be sent to.
	 *
	 * @return the destination hostID of the host that this message
	 * should be sent to.
	 */
    public HostID getDestHostID() {
		return dHostID;
	}
    
    /**
	 * Returns the destination agentID of the agent that this message
	 * should be sent to.
	 *
	 * @return the destination agentID of the agent that this message
	 * should be sent to.
	 */
    public AgentID getDestAgentID() {
		return dAgentID;
	}
	
	/**
	 * Sets the timeout flag within this reply.
	 */
	public void setTimedOut() {
		this.timedOut = true;
	}
	
	/**
	 * Returns true if the reply was sent due to a time out occuring.
	 */
	public boolean timedOut() {
		return timedOut;
	}
	
	/**
	 * Returns true if the request was rejected.
	 */
	public boolean isRejected() {
		return isRejected;
	}
	
	/**
	 * Returns the tuple resulting from the probing operation.
	 */
	public ETuple getTuple() {
		return tuple;
	}
	
	/**
	 * Returns the tuples resulting from the probing group operation.
	 */
	public ETuple[] getTuples() {
		return tuples;
	}
	
	/**
	 * Returns the agentID of the agent that sent this reply.
	 */
	public AgentID getSourceAgentID() {
		return sAgentID;
	}
	
	public String toString() {
		return "Reply";
	}
}

