package limone.io;

import java.io.*;
import eLights.*;
import limone.*;

/**
 * Defines a message requesting that an operation be performed.  This naturally
 * implements TwoWayMessage.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class Request implements TwoWayMessage, Serializable
{
	/**
	 * The type of request.  Possible
	 * values include Request.INP, Request.INGP, Request.RDP, Request.RDGP,
	 * Request.REGISTER_REACTIVE_PATTERN, Request.DEREGISTER_REACTIVE_PATTERN,
	 * Request.OUT.
	 */
	int type;
	
	/**
	 * The tuple that should be OUTed.
	 */
	ETuple tuple;
	
	/**
	 * The template that is used in the remote operation.
	 */
	ETemplate template;
	
	/**
	 * The reactive pattern that should be registered or deregistered.
	 */
	ReactivePattern rp;
	
	/**
	 * The AgentID of the agent that is sending this request.
	 */
	AgentID sAgentID;
	
	/**
	 * The AgentID of the agent that this request is destined to.
	 */
	AgentID dAgentID;
	
	/**
	 * The HostID of the LimoneServer on which the destination agent is
	 * residing on.
	 */
	HostID sHostID;
	
	/**
	 * The HostID of the LimoneServer on which the requesting agent is
	 * residing on.
	 */
	HostID dHostID;
	
	/**
	 * A unique integer that indentifies this request.  It is used to correlate
	 * the request with the response.
	 */
	int id;
	
	private static int enum = 0;
	
	/**
	 * The default constructor does not do anything.
	 */
	public Request() {
	}
	
	/**
	 * Stores the parameters as local variables.
	 *
	 * @param type the type of the operation that is being requested.
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param a the agent sending this request
	 */
	private Request(int type, AgentID destAgentID, HostID destHostID, Agent a) {
		this.type = type;
		this.dHostID = destHostID;
		this.dAgentID = destAgentID;
		this.sHostID = a.getHostID();
		this.sAgentID = a.getAgentID();
		id = enum++;
	}
	
	/**
	 * Creates a request to perform an INp, INgp,
	 * RDp, RDgp operation on a remote agent's tuple space.
	 *
	 * @param type the type of the operation that is being requested.  Possible
	 * values include Message.INP, Message.INGP, Message.RDP, and Message.RDGP
	 * @param template the template that is used to perform the operation
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param a the agent sending this request
	 */
	public Request(int type, AgentID destAgentID,
				   HostID destHostID, Agent a, ETemplate template) {
		this(type, destAgentID, destHostID, a);
		this.template = template;
		
	}
	
	/**
	 * Creates a request to perform a register reactive pattern
	 * and deregister reactive pattern.
	 *
	 * @param type the type of the operation that is being requested.  Possible
	 * values include Message.REGISTER_REACTIVE_PATTERN and
	 * Message.DEREGISTER_REACTIVE_PATTERN
	 * @param rp the reactive pattern that is being registered or deregistered
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param a the agent sending this request
	 */
	public Request(int type, AgentID destAgentID,
				   HostID destHostID, Agent a, ReactivePattern rp) {
		this(type, destAgentID, destHostID, a);
		this.rp = rp;
	}
	
	/**
	 * Creates a request to perform an OUT operation.
	 *
	 * @param tuple the tuple to place into the remote tuple space
	 * @param destAgentID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param destHostID the HostID of the LimoneServer on which the
	 * destination agent is residing on.
	 * @param a the agent sending this request
	 */
	public Request(AgentID destAgentID, HostID destHostID,
				   Agent a, ETuple tuple) {
		this(Message.OUT, destAgentID, destHostID, a);
		this.tuple = tuple;
	}
	
	/**
     * Returns the type of the message.  Type possible
     * types are defined within Message.
     *
     * @return the type of the message.  Type possible
     * types are defined within Message.
     */
    public int getType() {
		return type;
	}
	
	/**
     * Returns the destination hostID of the host that this message
     * should be sent to.
     *
     * @return the destination hostID of the host that this message
     * should be sent to.
     */
    public HostID getDestHostID() {
		return dHostID;
	}
    
    /**
     * Returns the destination agentID of the agent that this message
     * should be sent to.
     *
     * @return the destination agentID of the agent that this message
     * should be sent to.
     */
    public AgentID getDestAgentID(){
		return dAgentID;
	}
	
	/**
	 * Returns the hostID of the host that the reply should be sent.
	 */
	public HostID getReplyHostID() {
		return sHostID;
	}
	
	/**
     * Returns the agentID of the agent that the reply should be sent.
     */
	public AgentID getReplyAgentID(){
		return sAgentID;
	}
	
	/**
	 * Returns the agentID of the agent that sent this reply.
	 */
	public AgentID getSourceAgentID() {
		return sAgentID;
	}
	
	/**
	 * Returns the unique integer that indentifies this request.
	 * This integer is used to correlate the request with the response.
	 */
	public int getID() {
		return id;
	}
	
	/**
	 * Returns the tuple used in this request.
	 */
	public ETuple getTuple() {
		return tuple;
	}
	
	/**
	 * Returns the template used in this request.
	 */
	public ETemplate getTemplate() {
		return template;
	}
	
	/**
	 * Returns the unique enumeration value of this request.
	 */
	//public int getEnum() {
	//	return enum;
	//}
	
	/**
	 * Returns the reactive pattern that the agent is requesting to be
	 * registered.
	 */
	public ReactivePattern getRP() {
		return rp;
	}
	
	public String toString() {
		return "Request";
	}
}

