package limone.io;

import java.io.*;
import java.net.*;
import java.util.*;
import limone.*;

/**
 * Accepts clients connects and listens for messages of these clients.
 *
 * @author Chien-Liang Fok
 * @version 3/17/2003
 */
public class TCPMessageReceiver implements MessageReceiver,
    Runnable, Serializable {
    /**
	 * A reference to the LIMELiteServer using this receiver.
	 */
    private LimoneServer ls;
    
    /**
	 * The server socket.
	 */
    private ServerSocket ss;
    
    /**
	 * The thread that accepts incomming connections.
	 */
    private Thread acceptThread;
	
	/**
	 * The port being listened to.
	 */
	private int port;
    
    /**
	 * This message receiver listens for TCP connections
	 * and reads in commands from them.
	 *
	 * @param port the port on which to listen for connections
	 */
    public TCPMessageReceiver(int port){
		this.port = port;
		this.ls = LimoneServer.getServer();
		
		// create the server socket
		try{
			ss = new ServerSocket(port);
		}
		catch(Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
		
		// start the thread that accepts connections
		acceptThread = new Thread(this);
		acceptThread.start();
    }
    
    /**
	 * Stop the operation of the Message Receiver.
	 * Once the MessageReceiver has been terminated, it can no
	 * longer resume operation and should be discarded.
	 */
    public void kill() {
		
		try {
			ss.close();
			acceptThread.join();
		} catch(Exception e){
			e.printStackTrace();
		};
    }
	
	void log(String msg) {
		DebugLogger.log("TCPMessageReceiver", msg);
	}
    
    /**
	 * Sits in a loop waiting for clients to connect.  When a client connects,
	 * it creates a ClientHandler for it.
	 */
    public void run() {
		// continue to loop until an IOException is thrown
		try {
			while(true) {
				Socket s = null;
				log("Waiting for a connection on port " + port);
				if ((s = ss.accept()) != null) {
					log("Connection accepted, passing to client handler.");
					s.setTcpNoDelay(true);
					new ClientHandler(s);
				}
			}
		} catch(IOException e) {
			//if (!ss.isClosed())
			//	e.printStackTrace();
		} finally {
			try {
				ss.close();
			} catch(Exception e){}
		}
    }
    
    /**
	 * Handles incomming messages from a particular client.
	 */
    private class ClientHandler implements Runnable {
		/**
		 * The sock et to the client.
		 */
		Socket socket;
		/**
		 * The output streams.
		 */
		ObjectOutputStream out;
		/**
		 * The input streams.
		 */
		ObjectInputStream in;
		/**
		 * The thread that reads messages sent from the client.
		 */
		Thread chThread;
		/**
		 * Creates a clientHandler.
		 */
		public ClientHandler(Socket socket) {
			this.socket = socket;
			
			// extract the input and output streams
			try {
				// be sure to create the output stream before the input stream
				out = new ObjectOutputStream(socket.getOutputStream());
				in = new ObjectInputStream(socket.getInputStream());
			} catch(IOException e) {
				e.printStackTrace();
				return;
			}
			
			new Thread(this).start();
		}
		
		/**
		 * Sits in a loop listening for incomming messages.
		 */
		public void run() {
			try {
				log("Reading in object");
				Object o = in.readObject();
				
				if (o!= null && o instanceof Message) {
					Message msg = (Message)o;
					if(msg.getDestHostID().equals(ls.getHostID()))
						ls.newMsg(msg);
				}
			} catch(ClassNotFoundException e) {
				e.printStackTrace();
			} catch(IOException e) {
				e.printStackTrace();
			} finally {
				try {
					socket.close();
				} catch(Exception e) {}
			}
		}
		
		void log(String msg) {
			DebugLogger.log("TCPMessageReceiver: ClientHandler", msg);
		}
    }
}
