package limone.util;

import limone.*;
import limone.io.*;
import java.io.*;
import java.net.*;

/**
 * This class is used to load an agent onto a LimoneServer.  The
 * LimoneServer may either already exist, or be created prior to
 * loading the agent. If the "-newServer" option is used, then this class
 * creates a new LimoneServer.  If the "-newServer" option is not used,
 * this class loads the agent into an existing LimoneServer at the
 * specified port on the local machine.  By default, the -newServer flag is
 * set.
 *
 * <p>The following are the command line parameters.
 *
 *<TABLE>
 *<TR>
 *<TD><B>Option</B></TD>
 *<TD><B>Description</B></TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-debug </code> (default: <code>not set</code>) </TD>
 *<TD>Controls whether debug messages are printed to the screen.</TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-mPort {port}</code> (default: <code>6000</code>) </TD>
 *<TD>The multi-cast port used by this Limone Server.</TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-sPort {port}</code> (default: <code>2200</code>) </TD>
 *<TD>The single-cast port used by this Limone Server.</TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-mcastaddr {address}</code> (default: <code>230.1.2.3</code>) </TD>
 *<TD>The multicast group address. </TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-protocol {TCP|UDP}</code> (default: <code>TCP</code>) </TD>
 *<TD>Controls whether the single-cast protocol is UDP or TCP.</TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-beaconPeriod {period}</code> (default: <code>1000</code>) </TD>
 *<TD>The beacon period in milliseconds.</TD>
 *</TR>
 *

 *<TR>
 *<TD><code>-newServer</code> (default: <code>on</code>) </TD>
 *<TD>Controls wether a new server is created,.</TD>
 *</TR>
 *
 *<TR>
 *<TD><code>-load {agent class name}</code></TD>
 *<TD>The name of the agent's class to load onto the LimoneServer</TD>
 *</TR>
 *
 *
 *<TR>
 *<TD><code>-name {agent name}</code></TD>
 *<TD>The name of the agent to load onto the LimoneServer</TD>
 *</TR>
 *</TR>
 *</TABLE>
 *
 * @author Chien-Liang Fok
 * @version 5/13/2003
 */
public class Launcher implements Serializable, PropertyKeys {
    /**
	 * Determines whether this launcher creates a new LimoneServer
	 * or loads the agent into an existing one.  This is determined by
	 * the -newServer command line option.
	 */
    boolean newServer = false;
	
	/**
	 * The name of the class of the agent to be loaded.
	 */
	String agentClass = null;
	
	/**
	 * The user-defined name of the agent to be loaded, by default this
	 * is "Anonymous."
	 */
	String agentName = "Anonymous";
	
	/**
	 * The protocol used.
	 */
	String protocol = "TCP";
	
	/**
	 * The port of the local Limone server.
	 */
	int destHostPort = 2200;
	
    /**
	 *  Launches the LimeServer.
	 */
    public void launch(String[] args) {
		LimoneServer ls = LimoneServer.getServer();
		
		try {
			parseArgs(args);
			
			if (newServer && !ls.isBooted()) {
				ls.parseArgs(args);
				ls.boot();
			}
			
			if (agentClass != null) {
				//if (ls.isBooted())
				//	LimoneServer.getServer().loadAgent(agentClass, agentName);
				//else {
				// Get the local address.
				InetAddress addr = InetAddress.getLocalHost();
				
				// Create the message
				LoadAgentRequest lar = new LoadAgentRequest(
					new HostID(addr, destHostPort),
					agentClass,
					agentName);
				
				// Create the message sender depending on which protocol to use.
				MessageSender mSender = null;
				
				if (protocol.equals("TCP"))
					mSender =  new TCPMessageSender();
				else
					mSender =  new UDPMessageSender();
				
				// Send the message to the destination LimoneServer.
				mSender.sendMessage(lar);
				
				// kill the message sender
				mSender.kill();
			}
		} catch(Exception e) {
			e.printStackTrace();
			printHelp();
			System.exit(0);
		}
    }
	
	private void resetDefault() {
		newServer = false;
		agentClass = null;
		agentName = "Anonymous";
		protocol = "TCP";
		destHostPort = 2200;
	}
    
    private void parseArgs(String[] args) {
		if (args == null)
			throw new IllegalArgumentException();
		resetDefault();
		if (args.length != 0) {
			for (int i = 0; i < args.length; i++) {
				//System.out.println("current arg: " + args[i]);
				if (args[i].equals("-" + PROTOCOL_KEY))
					protocol = args[++i];
				else if (args[i].equals("-" + SINGLECAST_PORT_KEY))
					destHostPort = Integer.valueOf(args[++i]).intValue();
				else if (args[i].equals("-" + LOAD_KEY))
					agentClass = args[++i];
				else if (args[i].equals("-" + NAME_KEY))
					agentName = args[++i];
				else if (args[i].equals("-" + NEW_SERVER_KEY))
					newServer = true;
			}
		}
    }
    
    private void printHelp() {
		System.out.println(
			"\nUsage: java Limone.util.Launcher \n" +
				"                               "+
				"[-mPort <port>]\n"+
				"                               "+
				"[-sPort <port>]\n"+
				"                               "+
				"[-mcastaddr {224.0.0.0 to 239.255.255.255}]\n" +
				"                               "+
				"[-protocol {TCP or UDP, TCP default}]\n"+
				"                               "+
				"[-beaconPeriod <period in ms>]\n"+
				"                               "+
				"[-newServer]\n"+
				"                               "+
				"[-load <Agent.class> ] \n"+
				"                               "+
				"[-name {name} ]\n");
    }
    
    public static void main(String[] args) {
		(new Launcher()).launch(args);
    }
}

